# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import signal

from roboid.scanner import Scanner
from roboid.keyboard import Keyboard
from roboid.runner import Runner
from roboid.model import DeviceType
from roboid.model import DataType
from roboid.hamster import Hamster
from roboid.hamster_s import HamsterS
from roboid.turtle import Turtle
from roboid.albert_ai import AlbertAi

__version__ = "1.5.4"

__all__ = ["DeviceType", "DataType", "Hamster", "HamsterS", "Turtle", "AlbertAi", "Keyboard", "scan", "set_executable", "wait", "wait_until_ready", "wait_until", "when_do", "while_do", "parallel"]

def scan():
    Scanner.scan()

def set_executable(execute):
    Runner.set_executable(execute)

def wait(milliseconds):
    Runner.wait(milliseconds)

def wait_until_ready():
    Runner.wait_until_ready()

def wait_until(condition, args=None):
    Runner.wait_until(condition, args)

def when_do(condition, do, args=None):
    Runner.when_do(condition, do, args)

def while_do(condition, do, args=None):
    Runner.while_do(condition, do, args)

def parallel(*functions):
    Runner.parallel(functions)

def _handle_signal(signal, frame):
    Runner.shutdown()
    raise SystemExit

signal.signal(signal.SIGINT, _handle_signal)


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import time
import threading

from roboid.runner import Runner
from roboid.util import Util
from roboid.model import DeviceType
from roboid.model import DataType
from roboid.model import Roboid
from roboid.connector import Result
from roboid.hamster import Hamster
from roboid.serial_connector import SerialConnector


class HamsterConnectionChecker(object):
    def __init__(self, roboid):
        self._roboid = roboid

    def check(self, info):
        if info[1] == "Hamster" and info[2] == "04":
            self._roboid._set_model_code(0x04)
            return True
        elif info[2] == "0E":
            self._roboid._set_model_code(0x0E)
            return True
        else:
            return False


class HamsterRoboid(Roboid):
    _COLOR_TO_RGB = {
        0: (0, 0, 0),
        1: (0, 0, 255),
        2: (0, 255, 0),
        3: (0, 255, 255),
        4: (255, 0, 0),
        5: (255, 0, 255),
        6: (255, 255, 0),
        7: (255, 255, 255)
    }
    _SPEED_TO_GAIN = {
        1: 6,
        2: 6,
        3: 5,
        4: 5,
        5: 4,
        6: 4,
        7: 3,
        8: 3,
        9: 2,
        10: 2
    }

    def __init__(self, index):
        super(HamsterRoboid, self).__init__(Hamster.ID, "Hamster", 0x00400000)
        self._index = index
        self._connector = None
        self._ready = False
        self._thread_lock = threading.Lock()

        self._left_wheel = 0
        self._right_wheel = 0
        self._buzzer = 0
        self._output_a = 0
        self._output_b = 0
        self._topology = 0
        self._left_led = 0
        self._right_led = 0
        self._note = 0
        self._line_tracer_mode_written = False
        self._line_tracer_mode = 0
        self._line_tracer_speed = 5
        self._io_mode_a = 0
        self._io_mode_b = 0
        self._config_proximity = 2
        self._config_gravity = 0
        self._config_band_width = 3

        self._light = 0
        self._temperature = 0
        
        self._line_tracer_flag = 0
        self._line_tracer_event = 0
        self._line_tracer_state = 0
        self._line_tracer_count = 0
        
        self._event_tilt = -4
        self._event_battery_state = -1
        
        self._create_model()

    def _set_model_code(self, code):
        self._model_code = code

    def _create_model(self):
        from roboid.hamster import Hamster
        dict = self._device_dict = {}
        dict[Hamster.LEFT_WHEEL] = self._left_wheel_device = self._add_device(Hamster.LEFT_WHEEL, "LeftWheel", DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0)
        dict[Hamster.RIGHT_WHEEL] = self._right_wheel_device = self._add_device(Hamster.RIGHT_WHEEL, "RightWheel", DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0)
        dict[Hamster.BUZZER] = self._buzzer_device = self._add_device(Hamster.BUZZER, "Buzzer", DeviceType.EFFECTOR, DataType.FLOAT, 1, 0, 167772.15, 0.0)
        dict[Hamster.OUTPUT_A] = self._output_a_device = self._add_device(Hamster.OUTPUT_A, "OutputA", DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.OUTPUT_B] = self._output_b_device = self._add_device(Hamster.OUTPUT_B, "OutputB", DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.TOPOLOGY] = self._topology_device = self._add_device(Hamster.TOPOLOGY, "Topology", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.LEFT_LED] = self._left_led_device = self._add_device(Hamster.LEFT_LED, "LeftLed", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0)
        dict[Hamster.RIGHT_LED] = self._right_led_device = self._add_device(Hamster.RIGHT_LED, "RightLed", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0)
        dict[Hamster.NOTE] = self._note_device = self._add_device(Hamster.NOTE, "Note", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 88, 0)
        dict[Hamster.LINE_TRACER_MODE] = self._line_tracer_mode_device = self._add_device(Hamster.LINE_TRACER_MODE, "LineTracerMode", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.LINE_TRACER_SPEED] = self._line_tracer_speed_device = self._add_device(Hamster.LINE_TRACER_SPEED, "LineTracerSpeed", DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 5)
        dict[Hamster.IO_MODE_A] = self._io_mode_a_device = self._add_device(Hamster.IO_MODE_A, "IoModeA", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.IO_MODE_B] = self._io_mode_b_device = self._add_device(Hamster.IO_MODE_B, "IoModeB", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.CONFIG_PROXIMITY] = self._config_proximity_device = self._add_device(Hamster.CONFIG_PROXIMITY, "ConfigProximity", DeviceType.COMMAND, DataType.INTEGER, 1, 1, 7, 2)
        dict[Hamster.CONFIG_GRAVITY] = self._config_gravity_device = self._add_device(Hamster.CONFIG_GRAVITY, "ConfigGravity", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 3, 0)
        dict[Hamster.CONFIG_BAND_WIDTH] = self._config_band_width_device = self._add_device(Hamster.CONFIG_BAND_WIDTH, "ConfigBandWidth", DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 3)
        dict[Hamster.SIGNAL_STRENGTH] = self._signal_strength_device = self._add_device(Hamster.SIGNAL_STRENGTH, "SignalStrength", DeviceType.SENSOR, DataType.INTEGER, 1, -128, 0, 0)
        dict[Hamster.LEFT_PROXIMITY] = self._left_proximity_device = self._add_device(Hamster.LEFT_PROXIMITY, "LeftProximity", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.RIGHT_PROXIMITY] = self._right_proximity_device = self._add_device(Hamster.RIGHT_PROXIMITY, "RightProximity", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.LEFT_FLOOR] = self._left_floor_device = self._add_device(Hamster.LEFT_FLOOR, "LeftFloor", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.RIGHT_FLOOR] = self._right_floor_device = self._add_device(Hamster.RIGHT_FLOOR, "RightFloor", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.ACCELERATION] = self._acceleration_device = self._add_device(Hamster.ACCELERATION, "Acceleration", DeviceType.SENSOR, DataType.INTEGER, 3, -32768, 32767, 0)
        dict[Hamster.LIGHT] = self._light_device = self._add_device(Hamster.LIGHT, "Light", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 65535, 0)
        dict[Hamster.TEMPERATURE] = self._temperature_device = self._add_device(Hamster.TEMPERATURE, "Temperature", DeviceType.SENSOR, DataType.INTEGER, 1, -40, 88, 0)
        dict[Hamster.INPUT_A] = self._input_a_device = self._add_device(Hamster.INPUT_A, "InputA", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.INPUT_B] = self._input_b_device = self._add_device(Hamster.INPUT_B, "InputB", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.LINE_TRACER_STATE] = self._line_tracer_state_device = self._add_device(Hamster.LINE_TRACER_STATE, "LineTracerState", DeviceType.EVENT, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.TILT] = self._tilt_device = self._add_device(Hamster.TILT, "Tilt", DeviceType.EVENT, DataType.INTEGER, 1, -3, 3, 0)
        dict[Hamster.BATTERY_STATE] = self._battery_state_device = self._add_device(Hamster.BATTERY_STATE, "BatteryState", DeviceType.EVENT, DataType.INTEGER, 1, 0, 2, 2)

    def find_device_by_id(self, device_id):
        return self._device_dict.get(device_id)

    def _run(self):
        try:
            while self._running:
                if self._receive(self._connector):
                    self._send(self._connector)
                time.sleep(0.01)
        except:
            pass

    def _init(self, port_name=None):
        Runner.register_required()
        self._running = True
        thread = threading.Thread(target=self._run)
        self._thread = thread
        thread.daemon = True
        thread.start()

        tag = "Hamster[{}]".format(self._index)
        self._connector = SerialConnector(tag, HamsterConnectionChecker(self))
        result = self._connector.open(port_name)
        if result == Result.FOUND:
            while self._ready == False and self._is_disposed() == False:
                time.sleep(0.01)
        elif result == Result.NOT_AVAILABLE:
            Runner.register_checked()

    def _release(self):
        self._running = False
        thread = self._thread
        self._thread = None
        if thread:
            thread.join()

        connector = self._connector
        self._connector = None
        if connector:
            connector.close()

    def _dispose(self):
        if self._is_disposed() == False:
            super(HamsterRoboid, self)._dispose()
            self._release()

    def _reset(self):
        super(HamsterRoboid, self)._reset()

        self._left_wheel = 0
        self._right_wheel = 0
        self._buzzer = 0
        self._output_a = 0
        self._output_b = 0
        self._topology = 0
        self._left_led = 0
        self._right_led = 0
        self._note = 0
        self._line_tracer_mode_written = False
        self._line_tracer_mode = 0
        self._line_tracer_speed = 5
        self._io_mode_a = 0
        self._io_mode_b = 0
        self._config_proximity = 2
        self._config_gravity = 0
        self._config_band_width = 3

        self._light = 0
        self._temperature = 0
        
        self._line_tracer_event = 0
        self._line_tracer_state = 0
        self._line_tracer_count = 0
        
        self._event_tilt = -4
        self._event_battery_state = -1

    def _request_motoring_data(self):
        with self._thread_lock:
            self._left_wheel = self._left_wheel_device.read()
            self._right_wheel = self._right_wheel_device.read()
            self._buzzer = self._buzzer_device.read()
            self._output_a = self._output_a_device.read()
            self._output_b = self._output_b_device.read()
            if self._topology_device._is_written():
                self._topology = self._topology_device.read()
            if self._left_led_device._is_written():
                self._left_led = self._left_led_device.read()
            if self._right_led_device._is_written():
                self._right_led = self._right_led_device.read()
            if self._note_device._is_written():
                self._note = self._note_device.read()
            if self._line_tracer_mode_device._is_written():
                self._line_tracer_mode = self._line_tracer_mode_device.read()
                self._line_tracer_mode_written = True
            if self._line_tracer_speed_device._is_written():
                self._line_tracer_speed = self._line_tracer_speed_device.read()
            if self._io_mode_a_device._is_written():
                self._io_mode_a = self._io_mode_a_device.read()
            if self._io_mode_b_device._is_written():
                self._io_mode_b = self._io_mode_b_device.read()
            if self._config_proximity_device._is_written():
                self._config_proximity = self._config_proximity_device.read()
            if self._config_gravity_device._is_written():
                self._config_gravity = self._config_gravity_device.read()
            if self._config_band_width_device._is_written():
                self._config_band_width = self._config_band_width_device.read()
        self._clear_written()

    def _color_to_rgb(self, color):
        if isinstance(color, (int, float)):
            color = int(color)
            if color > 7: color = 7
            elif color < 0: color = 0
            return HamsterRoboid._COLOR_TO_RGB[color]
        return HamsterRoboid._COLOR_TO_RGB[0]

    def _speed_to_gain(self, speed):
        if isinstance(speed, (int, float)):
            speed = int(speed)
            if speed > 10: speed = 10
            elif speed < 1: speed = 1
            return HamsterRoboid._SPEED_TO_GAIN[speed]
        return 2

    def _encode_motoring_packet(self, address):
        result = ""
        with self._thread_lock:
            if self._model_code == 0x0E:
                result += "10"
                result += self._to_hex(self._left_wheel)
                result += self._to_hex(self._right_wheel)
                rgb = self._color_to_rgb(self._left_led)
                result += self._to_hex(rgb[0])
                result += self._to_hex(rgb[1])
                result += self._to_hex(rgb[2])
                rgb = self._color_to_rgb(self._right_led)
                result += self._to_hex(rgb[0])
                result += self._to_hex(rgb[1])
                result += self._to_hex(rgb[2])
                result += "000000"
                temp = self._line_tracer_mode & 0x0f
                if temp > 7: temp += 1
                if self._line_tracer_mode_written:
                    self._line_tracer_count = 0
                    if temp > 0:
                        self._line_tracer_flag = (self._line_tracer_flag % 15) + 1
                        self._line_tracer_event = 1
                    else:
                        self._line_tracer_event = 0
                    self._line_tracer_mode_written = False
                temp |= (self._line_tracer_flag & 0x0f) << 4
                result += self._to_hex(temp)
                temp = (self._line_tracer_speed & 0x0f) << 4
                temp |= self._speed_to_gain(self._line_tracer_speed) & 0x0f
                result += self._to_hex(temp)
                temp = (self._config_proximity & 0x07) << 5
                temp |= (self._config_band_width & 0x07) << 2
                temp |= self._config_gravity & 0x03
                result += self._to_hex(temp)
                temp = (self._io_mode_a & 0x0f) << 4
                temp |= self._io_mode_b & 0x0f
                result += self._to_hex(temp)
                result += self._to_hex(self._output_a)
                result += self._to_hex(self._output_b)
                if self._note > 0:
                    result += "01"
                    result += self._to_hex(self._note)
                else:
                    temp = self._buzzer
                    if temp > 6500: temp = 6500
                    result += self._to_hex2(Util.round(temp * 10) + 512)
            else:
                result += self._to_hex(self._topology & 0x0f)
                result += "0010"
                result += self._to_hex(self._left_wheel)
                result += self._to_hex(self._right_wheel)
                result += self._to_hex(self._left_led)
                result += self._to_hex(self._right_led)
                result += self._to_hex3(Util.round(self._buzzer * 100))
                result += self._to_hex(self._note)
                if self._line_tracer_mode_written:
                    if self._line_tracer_mode > 0:
                        self._line_tracer_flag ^= 0x80
                        self._line_tracer_event = 1
                    self._line_tracer_mode_written = False
                temp = (self._line_tracer_mode & 0x0f) << 3
                temp |= (self._line_tracer_speed - 1) & 0x07
                temp |= self._line_tracer_flag & 0x80
                result += self._to_hex(temp)
                result += self._to_hex(self._config_proximity)
                temp = (self._config_gravity & 0x0f) << 4
                temp |= self._config_band_width & 0x0f
                result += self._to_hex(temp)
                temp = (self._io_mode_a & 0x0f) << 4
                temp |= self._io_mode_b & 0x0f
                result += self._to_hex(temp)
                result += self._to_hex(self._output_a)
                result += self._to_hex(self._output_b)
                result += "000000"
        result += "-"
        result += address
        result += "\r"
        return result

    def _decode_sensory_packet(self, packet):
        packet = str(packet)
        if self._model_code == 0x0E:
            value = int(packet[0:1], 16)
            if value != 1: return False
            value = int(packet[6:8], 16)
            self._left_proximity_device._put(value)
            value = int(packet[8:10], 16)
            self._right_proximity_device._put(value)
            value2 = int(packet[38:40], 16)
            if (value2 & 0x01) == 0:
                self._light = int(packet[10:14], 16)
            else:
                value = int(packet[10:12], 16)
                if value > 0x7f: value -= 0x100
                self._temperature = Util.round(value / 2.0 + 23)
            self._light_device._put(self._light)
            self._temperature_device._put(self._temperature)
            value = int(packet[14:16], 16)
            self._left_floor_device._put(value)
            value = int(packet[16:18], 16)
            self._right_floor_device._put(value)
            acc_x = int(packet[18:22], 16)
            if acc_x > 0x7fff: acc_x -= 0x10000
            self._acceleration_device._put_at(0, acc_x)
            acc_y = int(packet[22:26], 16)
            if acc_y > 0x7fff: acc_y -= 0x10000
            self._acceleration_device._put_at(1, acc_y)
            acc_z = int(packet[26:30], 16)
            if acc_z > 0x7fff: acc_z -= 0x10000
            self._acceleration_device._put_at(2, acc_z)
            if acc_z < 8192 and acc_x > 8192 and acc_y > -4096 and acc_y < 4096: value = 1
            elif acc_z < 8192 and acc_x < -8192 and acc_y > -4096 and acc_y < 4096: value = -1
            elif acc_z < 8192 and acc_y > 8192 and acc_x > -4096 and acc_x < 4096: value = 2
            elif acc_z < 8192 and acc_y < -8192 and acc_x > -4096 and acc_x < 4096: value = -2
            elif acc_z > 12288 and acc_x > -8192 and acc_x < 8192 and acc_y > -8192 and acc_y < 8192: value = 3
            elif acc_z < -12288 and acc_x > -4096 and acc_x < 4096 and acc_y > -4096 and acc_y < 4096: value = -3
            else: value = 0
            if value != self._event_tilt:
                self._tilt_device._put(value, self._event_tilt != -4)
                self._event_tilt = value
            value = int(packet[30:32], 16)
            self._input_a_device._put(value)
            value = int(packet[32:34], 16)
            self._input_b_device._put(value)
            value = int(packet[36:38], 16)
            value -= 0x100
            self._signal_strength_device._put(value)
            value = (value2 >> 6) & 0x03
            if (value & 0x02) != 0:
                if self._line_tracer_event == 1:
                    if value == 0x02:
                        self._line_tracer_count += 1
                        if self._line_tracer_count > 5: self._line_tracer_event = 2
                    else:
                        self._line_tracer_event = 2
                if self._line_tracer_event == 2:
                    if value != self._line_tracer_state or self._line_tracer_count > 5:
                        self._line_tracer_state = value
                        self._line_tracer_state_device._put(value << 5)
                        if value == 0x02:
                            self._line_tracer_event = 0
                            self._line_tracer_count = 0
            value = (value2 >> 1) & 0x03
            if value == 0: value = 2
            elif value >= 2: value = 0
            if value != self._event_battery_state:
                self._battery_state_device._put(value, self._event_battery_state != -1)
                self._event_battery_state = value
        else:
            value = int(packet[4:5], 16)
            if value != 1: return False
            value = int(packet[6:8], 16)
            value -= 0x100
            self._signal_strength_device._put(value)
            value = int(packet[8:10], 16)
            self._left_proximity_device._put(value)
            value = int(packet[10:12], 16)
            self._right_proximity_device._put(value)
            value = int(packet[12:14], 16)
            self._left_floor_device._put(value)
            value = int(packet[14:16], 16)
            self._right_floor_device._put(value)
            acc_x = int(packet[16:20], 16)
            if acc_x > 0x7fff: acc_x -= 0x10000
            self._acceleration_device._put_at(0, acc_x)
            acc_y = int(packet[20:24], 16)
            if acc_y > 0x7fff: acc_y -= 0x10000
            self._acceleration_device._put_at(1, acc_y)
            acc_z = int(packet[24:28], 16)
            if acc_z > 0x7fff: acc_z -= 0x10000
            self._acceleration_device._put_at(2, acc_z)
            if acc_z < 8192 and acc_x > 8192 and acc_y > -4096 and acc_y < 4096: value = 1
            elif acc_z < 8192 and acc_x < -8192 and acc_y > -4096 and acc_y < 4096: value = -1
            elif acc_z < 8192 and acc_y > 8192 and acc_x > -4096 and acc_x < 4096: value = 2
            elif acc_z < 8192 and acc_y < -8192 and acc_x > -4096 and acc_x < 4096: value = -2
            elif acc_z > 12288 and acc_x > -8192 and acc_x < 8192 and acc_y > -8192 and acc_y < 8192: value = 3
            elif acc_z < -12288 and acc_x > -4096 and acc_x < 4096 and acc_y > -4096 and acc_y < 4096: value = -3
            else: value = 0
            if value != self._event_tilt:
                self._tilt_device._put(value, self._event_tilt != -4)
                self._event_tilt = value
            value = int(packet[28:30], 16)
            if value == 0:
                self._light = int(packet[30:34], 16)
            else:
                value = int(packet[30:32], 16)
                if value > 0x7f: value -= 0x100
                self._temperature = Util.round(value / 2.0 + 24)
                value = (int(packet[32:34], 16) + 200) / 100.0
                if value < 3.6: value = 0
                elif value <= 3.7: value = 1
                else: value = 2
                if value != self._event_battery_state:
                    self._battery_state_device._put(value, self._event_battery_state != -1)
                    self._event_battery_state = value
            self._light_device._put(self._light)
            self._temperature_device._put(self._temperature)
            value = int(packet[34:36], 16)
            self._input_a_device._put(value)
            value = int(packet[36:38], 16)
            self._input_b_device._put(value)
            value = int(packet[38:40], 16)
            if (value & 0x40) != 0:
                if self._line_tracer_event == 1:
                    if value != 0x40:
                        self._line_tracer_event = 2
                if self._line_tracer_event == 2:
                    if value != self._line_tracer_state:
                        self._line_tracer_state = value
                        self._line_tracer_state_device._put(value)
                        if value == 0x40:
                            self._line_tracer_event = 0
        return True

    def _receive(self, connector):
        if connector:
            packet = connector.read()
            if packet:
                if self._decode_sensory_packet(packet):
                    if self._ready == False:
                        self._ready = True
                        Runner.register_checked()
                    self._notify_sensory_device_data_changed()
                return True
        return False

    def _send(self, connector):
        if connector:
            packet = self._encode_motoring_packet(connector.get_address())
            connector.write(packet)


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA


class WriteQueue(object):
    def __init__(self, size):
        self.set_size(size)
        self._output = [0] * 19

    def set_size(self, size):
        self._buffer = bytearray(size)
        self._mask = size - 1
        self._provider = 0
        self._consumer = 0

    def reset(self):
        self._provider = 0
        self._consumer = 0

    def push(self, text, line):
        buffer = self._buffer
        mask = self._mask
        provider = self._provider
        consumer = self._consumer

        text = text.encode()
        length = len(text)
        if length > 0:
            for i in range(length):
                if ((provider - consumer) & mask) == mask:
                    consumer = (consumer + 1) & mask
                buffer[provider] = text[i]
                provider = (provider + 1) & mask
        if line:
            if ((provider - consumer) & mask) == mask:
                consumer = (consumer + 1) & mask
            buffer[provider] = ord("\r")
            provider = (provider + 1) & mask
        self._provider = provider
        self._consumer = consumer

    def pop(self):
        provider = self._provider
        consumer = self._consumer
        if provider == consumer: return None

        buffer = self._buffer
        mask = self._mask
        output = self._output
        length = (provider - consumer) & mask
        if length > 18: length = 18

        output[0] = length
        start = length + 1
        for i in range(1, length + 1):
            if provider == consumer:
                start = i
                break
            output[i] = buffer[consumer]
            consumer = (consumer + 1) & mask
        if start < 19:
            for i in range(start, 19):
                output[i] = 0
        self._consumer = consumer
        return output


class ReadQueue(object):
    def __init__(self, size):
        self.set_size(size)

    def set_size(self, size):
        self._buffer = bytearray(size)
        self._mask = size - 1
        self._provider = 0
        self._consumer = 0

    def reset(self):
        self._provider = 0
        self._consumer = 0

    def push(self, packet, offset=0):
        length = packet[offset]
        if length > 0:
            if length > 18: length = 18
            buffer = self._buffer
            mask = self._mask
            provider = self._provider
            consumer = self._consumer
            end = length + offset
            for i in range(1 + offset, end + 1):
                if ((provider - consumer) & mask) == mask:
                    consumer = (consumer + 1) & mask
                buffer[provider] = packet[i]
                provider = (provider + 1) & mask
            self._provider = provider
            self._consumer = consumer

    def pop(self, delimiter):
        provider = self._provider
        consumer = self._consumer
        if provider == consumer: return None

        buffer = self._buffer
        mask = self._mask
        if delimiter == 0:
            output = bytearray()
            while consumer != provider:
                output.append(buffer[consumer])
                consumer = (consumer + 1) & mask
            self._consumer = consumer
            return output.decode("utf-8")
        else:
            found = -1
            while consumer != provider:
                if buffer[consumer] == delimiter:
                    found = consumer
                    break
                consumer = (consumer + 1) & mask
            if found >= 0:
                consumer = self._consumer
                output = bytearray()
                while consumer != found:
                    output.append(buffer[consumer])
                    consumer = (consumer + 1) & mask
                self._consumer = (consumer + 1) & mask
                return output.decode("utf-8")
        return None


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA


DeviceType = type("Enum", (), {"SENSOR": 0, "EFFECTOR": 1, "EVENT": 2, "COMMAND": 3})
DataType = type("Enum", (), {"INTEGER": 4, "FLOAT": 5, "STRING": 6})


class NamedElement(object):
    def __init__(self, name):
        self.set_name(name)

    def get_name(self):
        return self._name

    def set_name(self, name):
        self._name = str(name)


class Device(NamedElement):
    def __init__(self, id, name, device_type, data_type, data_size, min_value, max_value, initial_value):
        super(Device, self).__init__(name)
        self._id = id & 0xfff00fff
        self._device_type = device_type
        self._data_type = data_type
        self._data_size = data_size
        self._min_value = min_value
        self._max_value = max_value
        self._initial_value = initial_value
        if data_size > 0:
            self._data = [initial_value] * data_size
        else:
            self._data = []
        self._event = False
        self._fired = False
        self._written = False
        self._can_notify = False
        self._device_data_changed_listeners = []

    def get_id(self):
        return self._id

    def get_device_type(self):
        return self._device_type

    def get_data_type(self):
        return self._data_type

    def get_data_size(self):
        return self._data_size

    def _is_written(self):
        return self._written

    def _clear_written(self):
        self._written = False

    def _check_data_type(self, value):
        if self._data_type == DataType.INTEGER or self._data_type == DataType.FLOAT:
            return isinstance(value, (int, float))
        elif self._data_type == DataType.STRING:
            return isinstance(value, str)
        return False

    def _check_range(self, value):
        if isinstance(value, (int, float)):
            if value < self._min_value:
                value = self._min_value
            elif value > self._max_value:
                value = self._max_value
            if self._data_type == DataType.INTEGER:
                return int(value)
            else:
                return float(value)
        return str(value)

    def e(self):
        return self._event

    def read(self, arg=None):
        this_data = self._data
        this_len = len(this_data)
        if isinstance(arg, (int, float)):
            index = int(arg)
            if index >= 0 and index < this_len:
                return this_data[index]
        elif isinstance(arg, list):
            length = min(this_len, len(arg))
            arg[:length] = this_data[:length]
            return length
        elif this_len > 0:
            return this_data[0]
        if self._data_type == DataType.STRING:
            return ""
        else:
            return 0

    def write(self, arg1=None, arg2=None):
        this_data = self._data
        if arg1 is None:
            self._fired = True
            self._written = True
            self._can_notify = True
        elif isinstance(arg1, (int, float)) and isinstance(arg2, (int, float, str)):
            if self._device_type == DeviceType.SENSOR or self._device_type == DeviceType.EVENT:
                return False
            index = int(arg1)
            if index < 0:
                return False
            if self._check_data_type(arg2) == False:
                return False
            this_len = len(this_data)
            if index >= this_len:
                if self._data_size < 0:
                    new_data = [self._initial_value] * (index + 1)
                    new_data[:this_len] = this_data
                    this_data = self._data = new_data
                else:
                    return False
            this_data[index] = self._check_range(arg2)
            self._fired = True
            self._written = True
            self._can_notify = True
            return True
        elif isinstance(arg1, (int, float, str)):
            if self._device_type == DeviceType.SENSOR or self._device_type == DeviceType.EVENT:
                return False
            if self._check_data_type(arg1) == False:
                return False
            if len(this_data) <= 0:
                if self._data_size < 0:
                    this_data = self._data = [0]
                else:
                    return False
            this_data[0] = self._check_range(arg1)
            self._fired = True
            self._written = True
            self._can_notify = True
            return True
        elif isinstance(arg1, (list, tuple)):
            if self._device_type == DeviceType.SENSOR or self._device_type == DeviceType.EVENT:
                return 0
            data_len = len(arg1)
            if data_len <= 0:
                return 0
            if self._data_size < 0:
                if len(this_data) != data_len:
                    this_data = self._data = [self._initial_value] * data_len
            this_len = len(this_data)
            if this_len <= 0:
                return 0
            length = min(this_len, data_len)
            written = False
            for i in range(length):
                if self._check_data_type(arg1[i]):
                    this_data[i] = self._check_range(arg1[i])
                    self._fired = True
                    self._written = True
                    self._can_notify = True
                    written = True
            if written:
                return length
        return 0

    def _put(self, value, fired=True):
        self._data[0] = value
        self._fired = fired
        self._can_notify = True

    def _put_at(self, index, value, fired=True):
        self._data[index] = value
        self._fired = fired
        self._can_notify = True

    def _put_empty(self, fired=True):
        self._fired = fired
        self._can_notify = True

    def _reset(self):
        self._event = False
        self._fired = False
        self._written = False
        if self._data_size > 0:
            self._data = [self._initial_value] * self._data_size
        else:
            self._data = []

    def add_device_data_changed_listener(self, listener):
        if listener:
            self._device_data_changed_listeners.append(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            self._device_data_changed_listeners.remove(listener)

    def clear_device_data_changed_listeners(self):
        self._device_data_changed_listeners = []

    def _update_device_state(self):
        self._event = self._fired
        self._fired = False

    def _notify_device_data_changed(self):
        if len(self._device_data_changed_listeners) > 0:
            can_notify = True
            if self._device_type == DeviceType.COMMAND or self._device_type == DeviceType.EVENT:
                can_notify = self._can_notify
            if can_notify:
                data = tuple(self._data)
                for listener in self._device_data_changed_listeners:
                    try:
                        listener.__func__(self, data)
                    except:
                        try:
                            listener(self, data)
                        except:
                            pass
        self._can_notify = False


class Roboid(NamedElement):
    def __init__(self, id, name, uid):
        super(Roboid, self).__init__(name)
        self._id = id
        self._uid = uid
        self._roboids = []
        self._devices = []
        self._sensory_devices = []
        self._motoring_devices = []
        self._disposed = False

    def _is_disposed(self):
        return self._disposed

    def _dispose(self):
        self._disposed = True
        self._reset()

    def _reset(self):
        for device in self._devices:
            device._reset()
        for roboid in self._roboids:
            roboid._reset()

    def _add_device(self, id, name, device_type, data_type, data_size, min_value, max_value, initial_value):
        dev = Device(id, name, device_type, data_type, data_size, min_value, max_value, initial_value)
        self._devices.append(dev)
        if device_type == DeviceType.SENSOR or device_type == DeviceType.EVENT:
            self._sensory_devices.append(dev)
        else:
            self._motoring_devices.append(dev)
        return dev

    def _add_roboid(self, roboid):
        if isinstance(roboid, Roboid):
            self._roboids.append(roboid)

    def get_id(self):
        return self._id

    def find_roboid_by_name(self, name):
        name = str(name).lower()
        dot = name.find(".")
        if dot < 0:
            for roboid in self._roboids:
                if name == roboid.get_name().lower():
                    return roboid
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self._roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_roboid_by_name(sub_name)
        return None

    def find_device_by_name(self, name):
        name = str(name).lower()
        dot = name.find(".")
        if dot < 0:
            for device in self._devices:
                if name == device.get_name().lower():
                    return device
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self._roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_device_by_name(sub_name)
        return None

    def find_device_by_id(self, device_id):
        uid = device_id & 0xfffff000
        if uid == self._uid:
            for device in self._devices:
                if device.get_id() == device_id:
                    return device
            return None

        for roboid in self._roboids:
            device = roboid.find_device_by_id(device_id)
            if device is not None:
                return device
        return None

    def add_device_data_changed_listener(self, listener):
        if listener:
            for device in self._devices:
                device.add_device_data_changed_listener(listener)
            for roboid in self._roboids:
                roboid.add_device_data_changed_listener(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            for device in self._devices:
                device.remove_device_data_changed_listener(listener)
            for roboid in self._roboids:
                roboid.remove_device_data_changed_listener(listener)

    def clear_device_data_changed_listeners(self):
        for device in self._devices:
            device.clear_device_data_changed_listeners()
        for roboid in self._roboids:
            roboid.clear_device_data_changed_listeners()

    def _to_hex(self, number):
        value = int(number)
        if value < 0: value += 0x100
        value = hex(value).upper()[2:]
        if len(value) > 1:
            return value
        return "0" + value

    def _to_hex2(self, number):
        value = int(number)
        if value < 0: value += 0x10000
        value = hex(value).upper()[2:]
        length = len(value)
        if length < 4:
            result = ""
            for i in range(length, 4):
                result += "0"
            return result + value
        return value

    def _to_hex3(self, number):
        value = int(number)
        if value < 0: value += 0x1000000
        value = hex(value).upper()[2:]
        length = len(value)
        if length < 6:
            result = ""
            for i in range(length, 6):
                result += "0"
            return result + value
        return value

    def _request_motoring_data(self):
        pass

    def _update_sensory_device_state(self):
        for device in self._sensory_devices:
            device._update_device_state()
        for roboid in self._roboids:
            roboid._update_sensory_device_state()

    def _update_motoring_device_state(self):
        for device in self._motoring_devices:
            device._update_device_state()
        for roboid in self._roboids:
            roboid._update_motoring_device_state()

    def _notify_sensory_device_data_changed(self):
        for device in self._sensory_devices:
            device._notify_device_data_changed()
        for roboid in self._roboids:
            roboid._notify_sensory_device_data_changed()

    def _notify_motoring_device_data_changed(self):
        for device in self._motoring_devices:
            device._notify_device_data_changed()
        for roboid in self._roboids:
            roboid._notify_motoring_device_data_changed()

    def _clear_written(self):
        for device in self._devices:
            device._clear_written()
        for roboid in self._roboids:
            roboid._clear_written()


class Robot(NamedElement):
    def __init__(self, id, name, index):
        super(Robot, self).__init__(name)
        self._id = id
        self._index = index
        self._roboids = []

    def dispose(self):
        for roboid in self._roboids:
            roboid._dispose()

    def get_id(self):
        return self._id

    def get_index(self):
        return self._index

    def reset(self):
        for roboid in self._roboids:
            roboid._reset()

    def _add_roboid(self, roboid):
        if isinstance(roboid, Roboid):
            self._roboids.append(roboid)

    def find_roboid_by_name(self, name):
        name = str(name).lower()
        dot = name.find(".")
        if dot < 0:
            for roboid in self._roboids:
                if name == roboid.get_name().lower():
                    return roboid
        else:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self._roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_roboid_by_name(sub_name)
        return None

    def find_device_by_name(self, name):
        name = str(name).lower()
        dot = name.find(".")
        if dot >= 0:
            roboid_name = name[:dot]
            sub_name = name[dot+1:]
            for roboid in self._roboids:
                if roboid_name == roboid.get_name().lower():
                    return roboid.find_device_by_name(sub_name)
        return None

    def find_device_by_id(self, device_id):
        for roboid in self._roboids:
            device = roboid.find_device_by_id(device_id)
            if device is not None:
                return device
        return None

    def e(self, device_id):
        device = self.find_device_by_id(device_id)
        if device is None: return False
        return device.e()

    def read(self, device_id, arg=None):
        device = self.find_device_by_id(device_id)
        if device is None: return 0
        return device.read(arg)

    def write(self, device_id, arg1, arg2=None):
        device = self.find_device_by_id(device_id)
        if device is None:
            if isinstance(arg1, (int, float)) and isinstance(arg2, (int, float, str)):
                return False
            elif isinstance(arg1, (int, float, str)):
                return False
            return 0
        return device.write(arg1, arg2)

    def add_device_data_changed_listener(self, listener):
        if listener:
            for roboid in self._roboids:
                roboid.add_device_data_changed_listener(listener)

    def remove_device_data_changed_listener(self, listener):
        if listener:
            for roboid in self._roboids:
                roboid.remove_device_data_changed_listener(listener)

    def clear_device_data_changed_listeners(self):
        for roboid in self._roboids:
            roboid.clear_device_data_changed_listeners()

    def _request_motoring_data(self):
        for roboid in self._roboids:
            roboid._request_motoring_data()

    def _update_sensory_device_state(self):
        for roboid in self._roboids:
            roboid._update_sensory_device_state()

    def _update_motoring_device_state(self):
        for roboid in self._roboids:
            roboid._update_motoring_device_state()

    def _notify_sensory_device_data_changed(self):
        for roboid in self._roboids:
            roboid._notify_sensory_device_data_changed()

    def _notify_motoring_device_data_changed(self):
        for roboid in self._roboids:
            roboid._notify_motoring_device_data_changed()


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import os


if os.name == "nt":  # sys.platform == "win32":
    import msvcrt

    class Keyboard(object):
        BACKSPACE = 8
        TAB = 9
        ENTER = 13
        ESC = 27
        F1 = 59
        F2 = 60
        F3 = 61
        F4 = 62
        F5 = 63
        F6 = 64
        F7 = 65
        F8 = 66
        F9 = 67
        F10 = 68
        F11 = 133
        F12 = 134

        HOME = 71
        UP = 72
        PAGE_UP = 73
        LEFT = 75
        RIGHT = 77
        END = 79
        DOWN = 80
        PAGE_DOWN = 81
        INSERT = 82
        DELETE = 83
        
        _special_keys = (BACKSPACE, TAB, ENTER, ESC)
        _function_numpads = (F1, F2, F3, F4, F5, F6, F7, F8, F9, F10, HOME, UP, PAGE_UP, LEFT, RIGHT, END, DOWN, PAGE_DOWN, INSERT, DELETE)

        @staticmethod
        def read():
            if msvcrt.kbhit():
                key = msvcrt.getch()
                code = ord(key)
                if code == 224: # special key (F11, F12, HOME, UP, PAGE_UP, LEFT, RIGHT, END, DOWN, PAGE_DOWN, INSERT, DELETE)
                    return ord(msvcrt.getch())
                elif code == 0: # function key (F1 - F10) or num pad
                    code = ord(msvcrt.getch())
                    if code in Keyboard._function_numpads:
                        return code
                elif code in Keyboard._special_keys:
                    return code
                else:
                    return key.decode("utf-8")
else:
    import sys
    import termios
    from contextlib import contextmanager

    class Keyboard(object):
        BACKSPACE = 127
        TAB = 9
        ENTER = 10
        F1 = 80 # 27 79 80
        F2 = 81 # 27 79 81
        F3 = 82 # 27 79 82
        F4 = 83 # 27 79 83
        F5 = 53 # 27 91 49 53 126
        F6 = 55 # 27 91 49 55 126
        F7 = 56 # 27 91 49 56 126
        F8 = 57 # 27 91 49 57 126
        F9 = 48 # 27 91 50 48 126
        F10 = 49 # 27 91 50 49 126
        F11 = 51
        F12 = 52 # 27 91 50 52 126

        HOME = 72 # 27 91 72
        UP = 65 # 27 91 65
        PAGE_UP = 153 # 27 91 53 126
        LEFT = 68 # 27 91 68
        RIGHT = 67 # 27 91 67
        END = 70 # 27 91 70
        DOWN = 66 # 27 91 66
        PAGE_DOWN = 154 # 27 91 54 126
        INSERT = 150 # 27 91 50 126
        DELETE = 151 # 27 91 51 126
        
        _special_keys = (BACKSPACE, TAB, ENTER)

        @staticmethod
        @contextmanager
        def _mode(file):
            old_attrs = termios.tcgetattr(file.fileno())
            new_attrs = old_attrs[:]
            new_attrs[3] = new_attrs[3] & ~(termios.ECHO | termios.ICANON)
            try:
                termios.tcsetattr(file.fileno(), termios.TCSADRAIN, new_attrs)
                yield
            finally:
                termios.tcsetattr(file.fileno(), termios.TCSADRAIN, old_attrs)

        @staticmethod
        def read():
            with Keyboard._mode(sys.stdin):
                key = sys.stdin.read(1)
                code = ord(key)
                if code == 27: # special key
                    code = ord(sys.stdin.read(1))
                    if code == 79:
                        return ord(sys.stdin.read(1))
                    elif code == 91:
                        code = ord(sys.stdin.read(1))
                        if code >= 65:
                            return code
                        else:
                            code2 = ord(sys.stdin.read(1))
                            if code2 == 126:
                                return code + 100
                            else:
                                sys.stdin.read(1) # 126
                                return code2
                    else:
                        return code
                elif code in Keyboard._special_keys:
                    return code
                else:
                    return key


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import time
import threading
from timeit import default_timer as timer

from roboid.runner import Runner
from roboid.util import Util
from roboid.model import DeviceType
from roboid.model import DataType
from roboid.model import Roboid
from roboid.connector import Result
from roboid.albert_ai import AlbertAi
from roboid.serial_connector import SerialConnector


class AlbertAiConnectionChecker(object):
    def __init__(self, roboid):
        self._roboid = roboid

    def check(self, info):
        return info[1] == "Albert AI" and info[2] == "0A"


class AlbertAiTouchChecker(object):
    def __init__(self):
        self._down = False
        self._clicked = False
        self._long_pressed = False
        self._long_long_pressed = False
        self._check_long_pressed = False
        self._check_long_long_pressed = False
        self._pressed_time = 0

    def reset(self):
        self._down = False
        self._clicked = False
        self._long_pressed = False
        self._long_long_pressed = False
        self._check_long_pressed = False
        self._check_long_long_pressed = False
        self._pressed_time = 0

    def is_clicked(self):
        return self._clicked

    def is_long_pressed(self):
        return self._long_pressed

    def is_long_long_pressed(self):
        return self._long_long_pressed

    def check(self, touched):
        if self._down:
            if touched:
                self._clicked = False
                if self._check_long_pressed:
                    if timer() - self._pressed_time > 1.5:
                        self._check_long_pressed = False
                        self._long_pressed = True
                if self._check_long_long_pressed:
                    if timer() - self._pressed_time > 3:
                        self._check_long_long_pressed = False
                        self._long_long_pressed = True
            else:
                if timer() - self._pressed_time < 0.75:
                    self._clicked = True
                self._long_pressed = False
                self._long_long_pressed = False
                self._check_long_pressed = False
                self._check_long_long_pressed = False
        else:
            self._clicked = False
            self._long_pressed = False
            self._long_long_pressed = False
            if touched:
                self._check_long_pressed = True
                self._check_long_long_pressed = True
                self._pressed_time = timer()
            else:
                self._check_long_pressed = False
                self._check_long_long_pressed = False
        self._down = touched


class AlbertAiRoboid(Roboid):
    _SOUNDS = {
        AlbertAi.SOUND_OFF: 0x00,
        AlbertAi.SOUND_BEEP: 0x01,
        AlbertAi.SOUND_RANDOM_BEEP: 0x05,
        AlbertAi.SOUND_NOISE: 0x07,
        AlbertAi.SOUND_SIREN: 0x10,
        AlbertAi.SOUND_ENGINE: 0x20,
        AlbertAi.SOUND_ROBOT: 0x30
    }

    def __init__(self, index):
        super(AlbertAiRoboid, self).__init__(AlbertAi.ID, "AlbertAi", 0x00a00000)
        self._index = index
        self._connector = None
        self._ready = False
        self._thread_lock = threading.Lock()

        self._left_wheel = 0
        self._right_wheel = 0
        self._left_eye_red = 0
        self._left_eye_green = 0
        self._left_eye_blue = 0
        self._right_eye_red = 0
        self._right_eye_green = 0
        self._right_eye_blue = 0
        self._mic_led_red = 0
        self._mic_led_green = 0
        self._mic_led_blue = 0
        self._buzzer = 0
        self._pulse = 0
        self._note = 0
        self._sound = 0
        self._board_width = 0
        self._board_height = 0

        self._acceleration_x = [0] * 10
        self._acceleration_y = [0] * 10
        self._acceleration_z = [0] * 10
        self._acceleration_sum_x = 0.0
        self._acceleration_sum_y = 0.0
        self._acceleration_sum_z = 0.0
        self._acceleration_index = 0
        self._acceleration_count = 0

        self._mic_checker = AlbertAiTouchChecker()
        self._volume_up_checker = AlbertAiTouchChecker()
        self._volume_down_checker = AlbertAiTouchChecker()
        self._play_checker = AlbertAiTouchChecker()
        self._back_checker = AlbertAiTouchChecker()

        self._wheel_id = 0
        self._wheel_pulse = 0
        self._wheel_pulse_prev = -1
        self._wheel_event = 0
        self._wheel_state = 0
        self._wheel_count = 0
        self._wheel_move = False
        self._wheel_move_count = 0
        self._pulse_written = False

        self._current_sound = 0
        self._sound_repeat = 1
        self._sound_flag = 0
        self._sound_event = 0
        self._sound_state = 0
        self._sound_count = 0
        self._sound_written = False

        self._event_mic_touch = -1
        self._event_volume_up_touch = -1
        self._event_volume_down_touch = -1
        self._event_play_touch = -1
        self._event_back_touch = -1
        self._event_mic_clicked = False
        self._event_volume_up_clicked = False
        self._event_volume_down_clicked = False
        self._event_play_clicked = False
        self._event_back_clicked = False
        self._event_mic_long_pressed = False
        self._event_volume_up_long_pressed = False
        self._event_volume_down_long_pressed = False
        self._event_play_long_pressed = False
        self._event_back_long_pressed = False
        self._event_mic_long_long_pressed = False
        self._event_volume_up_long_long_pressed = False
        self._event_volume_down_long_long_pressed = False
        self._event_play_long_long_pressed = False
        self._event_back_long_long_pressed = False
        self._event_tap_id = -1
        self._event_oid_mode_id = -1
        self._event_oid = -2
        self._event_lift = -1
        self._event_pulse_count = -1
        self._event_battery_state = -1
        self._event_tilt = -4
        
        self._create_model()

    def _create_model(self):
        from roboid.albert_ai import AlbertAi
        dict = self._device_dict = {}
        dict[AlbertAi.SPEAKER] = self._speaker_device = self._add_device(AlbertAi.SPEAKER, "Speaker", DeviceType.EFFECTOR, DataType.INTEGER, 480, -32768, 32767, 0)
        dict[AlbertAi.VOLUME] = self._volume_device = self._add_device(AlbertAi.VOLUME, "Volume", DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 300, 100)
        dict[AlbertAi.LIP] = self._lip_device = self._add_device(AlbertAi.LIP, "Lip", DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 100, 0)
        dict[AlbertAi.LEFT_WHEEL] = self._left_wheel_device = self._add_device(AlbertAi.LEFT_WHEEL, "LeftWheel", DeviceType.EFFECTOR, DataType.FLOAT, 1, -100, 100, 0)
        dict[AlbertAi.RIGHT_WHEEL] = self._right_wheel_device = self._add_device(AlbertAi.RIGHT_WHEEL, "RightWheel", DeviceType.EFFECTOR, DataType.FLOAT, 1, -100, 100, 0)
        dict[AlbertAi.LEFT_EYE] = self._left_eye_device = self._add_device(AlbertAi.LEFT_EYE, "LeftEye", DeviceType.EFFECTOR, DataType.INTEGER, 3, 0, 255, 0)
        dict[AlbertAi.RIGHT_EYE] = self._right_eye_device = self._add_device(AlbertAi.RIGHT_EYE, "RightEye", DeviceType.EFFECTOR, DataType.INTEGER, 3, 0, 255, 0)
        dict[AlbertAi.MIC_LED] = self._mic_led_device = self._add_device(AlbertAi.MIC_LED, "MicLed", DeviceType.EFFECTOR, DataType.INTEGER, 3, 0, 255, 0)
        dict[AlbertAi.BUZZER] = self._buzzer_device = self._add_device(AlbertAi.BUZZER, "Buzzer", DeviceType.EFFECTOR, DataType.FLOAT, 1, 0.0, 6553.5, 0.0)
        dict[AlbertAi.PULSE] = self._pulse_device = self._add_device(AlbertAi.PULSE, "Pulse", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 65535, 0)
        dict[AlbertAi.NOTE] = self._note_device = self._add_device(AlbertAi.NOTE, "Note", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 88, 0)
        dict[AlbertAi.SOUND] = self._sound_device = self._add_device(AlbertAi.SOUND, "Sound", DeviceType.COMMAND, DataType.INTEGER, 1, 0, 127, 0)
        dict[AlbertAi.BOARD_SIZE] = self._board_size_device = self._add_device(AlbertAi.BOARD_SIZE, "BoardSize", DeviceType.COMMAND,DataType.INTEGER, 2, 0, 268435455, 0)
        dict[AlbertAi.SIGNAL_STRENGTH] = self._signal_strength_device = self._add_device(AlbertAi.SIGNAL_STRENGTH, "SignalStrength", DeviceType.SENSOR, DataType.INTEGER, 1, -128, 0, 0)
        dict[AlbertAi.LEFT_PROXIMITY] = self._left_proximity_device = self._add_device(AlbertAi.LEFT_PROXIMITY, "LeftProximity", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[AlbertAi.RIGHT_PROXIMITY] = self._right_proximity_device = self._add_device(AlbertAi.RIGHT_PROXIMITY, "RightProximity", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[AlbertAi.ACCELERATION] = self._acceleration_device = self._add_device(AlbertAi.ACCELERATION, "Acceleration", DeviceType.SENSOR, DataType.INTEGER, 3, -8192, 8191, 0)
        dict[AlbertAi.POSITION] = self._position_device = self._add_device(AlbertAi.POSITION, "Position", DeviceType.SENSOR, DataType.INTEGER, 2, -1, 268435454, -1)
        dict[AlbertAi.ORIENTATION] = self._orientation_device = self._add_device(AlbertAi.ORIENTATION, "Orientation", DeviceType.SENSOR, DataType.INTEGER, 1, -200, 180, -200)
        dict[AlbertAi.LIGHT] = self._light_device = self._add_device(AlbertAi.LIGHT, "Light", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 65535, 0)
        dict[AlbertAi.BATTERY] = self._battery_device = self._add_device(AlbertAi.BATTERY, "Battery", DeviceType.SENSOR, DataType.INTEGER, 1, 0, 100, 0)
        dict[AlbertAi.MIC_TOUCH] = self._mic_touch_device = self._add_device(AlbertAi.MIC_TOUCH, "MicTouch", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.VOLUME_UP_TOUCH] = self._volume_up_touch_device = self._add_device(AlbertAi.VOLUME_UP_TOUCH, "VolumeUpTouch", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.VOLUME_DOWN_TOUCH] = self._volume_down_touch_device = self._add_device(AlbertAi.VOLUME_DOWN_TOUCH, "VolumeDownTouch", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.PLAY_TOUCH] = self._play_touch_device = self._add_device(AlbertAi.PLAY_TOUCH, "PlayTouch", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.BACK_TOUCH] = self._back_touch_device = self._add_device(AlbertAi.BACK_TOUCH, "BackTouch", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.MIC_CLICKED] = self._mic_clicked_device = self._add_device(AlbertAi.MIC_CLICKED, "MicClicked", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_UP_CLICKED] = self._volume_up_clicked_device = self._add_device(AlbertAi.VOLUME_UP_CLICKED, "VolumeUpClicked", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_DOWN_CLICKED] = self._volume_down_clicked_device = self._add_device(AlbertAi.VOLUME_DOWN_CLICKED, "VolumeDownClicked", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.PLAY_CLICKED] = self._play_clicked_device = self._add_device(AlbertAi.PLAY_CLICKED, "PlayClicked", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.BACK_CLICKED] = self._back_clicked_device = self._add_device(AlbertAi.BACK_CLICKED, "BackClicked", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.MIC_LONG_PRESSED] = self._mic_long_pressed_device = self._add_device(AlbertAi.MIC_LONG_PRESSED, "MicLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_UP_LONG_PRESSED] = self._volume_up_long_pressed_device = self._add_device(AlbertAi.VOLUME_UP_LONG_PRESSED, "VolumeUpLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_DOWN_LONG_PRESSED] = self._volume_down_long_pressed_device = self._add_device(AlbertAi.VOLUME_DOWN_LONG_PRESSED, "VolumeDownLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.PLAY_LONG_PRESSED] = self._play_long_pressed_device = self._add_device(AlbertAi.PLAY_LONG_PRESSED, "PlayLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.BACK_LONG_PRESSED] = self._back_long_pressed_device = self._add_device(AlbertAi.BACK_LONG_PRESSED, "BackLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.MIC_LONG_LONG_PRESSED] = self._mic_long_long_pressed_device = self._add_device(AlbertAi.MIC_LONG_LONG_PRESSED, "MicLongLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_UP_LONG_LONG_PRESSED] = self._volume_up_long_long_pressed_device = self._add_device(AlbertAi.VOLUME_UP_LONG_LONG_PRESSED, "VolumeUpLongLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.VOLUME_DOWN_LONG_LONG_PRESSED] = self._volume_down_long_long_pressed_device = self._add_device(AlbertAi.VOLUME_DOWN_LONG_LONG_PRESSED, "VolumeDownLongLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.PLAY_LONG_LONG_PRESSED] = self._play_long_long_pressed_device = self._add_device(AlbertAi.PLAY_LONG_LONG_PRESSED, "PlayLongLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.BACK_LONG_LONG_PRESSED] = self._back_long_long_pressed_device = self._add_device(AlbertAi.BACK_LONG_LONG_PRESSED, "BackLongLongPressed", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.TAP] = self._tap_device = self._add_device(AlbertAi.TAP, "Tap", DeviceType.EVENT, DataType.INTEGER, 0, 0, 0, 0)
        dict[AlbertAi.OID_MODE] = self._oid_mode_device = self._add_device(AlbertAi.OID_MODE, "OidMode", DeviceType.EVENT, DataType.INTEGER, 1, 0, 15, 0)
        dict[AlbertAi.OID] = self._oid_device = self._add_device(AlbertAi.OID, "Oid", DeviceType.EVENT, DataType.INTEGER, 1, -1, 268435455, -1)
        dict[AlbertAi.LIFT] = self._lift_device = self._add_device(AlbertAi.LIFT, "Lift", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.PULSE_COUNT] = self._pulse_count_device = self._add_device(AlbertAi.PULSE_COUNT, "PulseCount", DeviceType.EVENT, DataType.INTEGER, 1, 0, 65535, 0)
        dict[AlbertAi.WHEEL_STATE] = self._wheel_state_device = self._add_device(AlbertAi.WHEEL_STATE, "WheelState", DeviceType.EVENT, DataType.INTEGER, 1, 0, 3, 0)
        dict[AlbertAi.SOUND_STATE] = self._sound_state_device = self._add_device(AlbertAi.SOUND_STATE, "SoundState", DeviceType.EVENT, DataType.INTEGER, 1, 0, 1, 0)
        dict[AlbertAi.BATTERY_STATE] = self._battery_state_device = self._add_device(AlbertAi.BATTERY_STATE, "BatteryState", DeviceType.EVENT, DataType.INTEGER, 1, 0, 2, 2)
        dict[AlbertAi.TILT] = self._tilt_device = self._add_device(AlbertAi.TILT, "Tilt", DeviceType.EVENT, DataType.INTEGER, 1, -3, 3, 0)

    def find_device_by_id(self, device_id):
        return self._device_dict.get(device_id)

    def _run(self):
        try:
            while self._running:
                if self._receive(self._connector):
                    self._send(self._connector)
                time.sleep(0.01)
        except:
            pass

    def _init(self, port_name=None):
        Runner.register_required()
        self._running = True
        thread = threading.Thread(target=self._run)
        self._thread = thread
        thread.daemon = True
        thread.start()

        tag = "AlbertAi[{}]".format(self._index)
        self._connector = SerialConnector(tag, AlbertAiConnectionChecker(self))
        result = self._connector.open(port_name)
        if result == Result.FOUND:
            while self._ready == False and self._is_disposed() == False:
                time.sleep(0.01)
        elif result == Result.NOT_AVAILABLE:
            Runner.register_checked()

    def _release(self):
        self._running = False
        thread = self._thread
        self._thread = None
        if thread:
            thread.join()

        connector = self._connector
        self._connector = None
        if connector:
            connector.close()

    def _dispose(self):
        if self._is_disposed() == False:
            super(AlbertAiRoboid, self)._dispose()
            self._release()

    def _reset(self):
        super(AlbertAiRoboid, self)._reset()

        self._left_wheel = 0
        self._right_wheel = 0
        self._left_eye_red = 0
        self._left_eye_green = 0
        self._left_eye_blue = 0
        self._right_eye_red = 0
        self._right_eye_green = 0
        self._right_eye_blue = 0
        self._mic_led_red = 0
        self._mic_led_green = 0
        self._mic_led_blue = 0
        self._buzzer = 0
        self._pulse = 0
        self._note = 0
        self._sound = 0
        self._board_width = 0
        self._board_height = 0

        self._acceleration_sum_x = 0.0
        self._acceleration_sum_y = 0.0
        self._acceleration_sum_z = 0.0
        self._acceleration_index = 0
        self._acceleration_count = 0

        self._mic_checker.reset()
        self._volume_up_checker.reset()
        self._volume_down_checker.reset()
        self._play_checker.reset()
        self._back_checker.reset()

        self._wheel_pulse = 0
        self._wheel_pulse_prev = -1
        self._wheel_event = 0
        self._wheel_state = 0
        self._wheel_count = 0
        self._wheel_move = False
        self._wheel_move_count = 0
        self._pulse_written = False

        self._current_sound = 0
        self._sound_repeat = 1
        self._sound_event = 0
        self._sound_state = 0
        self._sound_count = 0
        self._sound_written = False

        self._event_mic_touch = -1
        self._event_volume_up_touch = -1
        self._event_volume_down_touch = -1
        self._event_play_touch = -1
        self._event_back_touch = -1
        self._event_mic_clicked = False
        self._event_volume_up_clicked = False
        self._event_volume_down_clicked = False
        self._event_play_clicked = False
        self._event_back_clicked = False
        self._event_mic_long_pressed = False
        self._event_volume_up_long_pressed = False
        self._event_volume_down_long_pressed = False
        self._event_play_long_pressed = False
        self._event_back_long_pressed = False
        self._event_mic_long_long_pressed = False
        self._event_volume_up_long_long_pressed = False
        self._event_volume_down_long_long_pressed = False
        self._event_play_long_long_pressed = False
        self._event_back_long_long_pressed = False
        self._event_tap_id = -1
        self._event_oid_mode_id = -1
        self._event_oid = -2
        self._event_lift = -1
        self._event_pulse_count = -1
        self._event_battery_state = -1
        self._event_tilt = -4

    def _request_motoring_data(self):
        with self._thread_lock:
            self._left_wheel = self._left_wheel_device.read()
            self._right_wheel = self._right_wheel_device.read()
            self._left_eye_red = self._left_eye_device.read(0)
            self._left_eye_green = self._left_eye_device.read(1)
            self._left_eye_blue = self._left_eye_device.read(2)
            self._right_eye_red = self._right_eye_device.read(0)
            self._right_eye_green = self._right_eye_device.read(1)
            self._right_eye_blue = self._right_eye_device.read(2)
            self._mic_led_red = self._mic_led_device.read(0)
            self._mic_led_green = self._mic_led_device.read(1)
            self._mic_led_blue = self._mic_led_device.read(2)
            self._buzzer = self._buzzer_device.read()
            if self._pulse_device._is_written():
                self._pulse = self._pulse_device.read()
                self._pulse_written = True
            if self._note_device._is_written():
                self._note = self._note_device.read()
            if self._sound_device._is_written():
                self._sound = self._sound_device.read()
                self._sound_written = True
            if self._board_size_device._is_written():
                self._board_width = self._board_size_device.read(0)
                self._board_height = self._board_size_device.read(1)
        self._clear_written()

    def _get_sound(self, sound):
        if isinstance(sound, (int, float)):
            sound = int(sound)
            if sound in AlbertAiRoboid._SOUNDS:
                return AlbertAiRoboid._SOUNDS[sound]
        return 0

    def _run_sound(self, sound, repeat):
        if isinstance(sound, (int, float)) and isinstance(repeat, (int, float)):
            sound = int(sound)
            repeat = int(repeat)
            if repeat < 0: repeat = -1
            if repeat != 0:
                self._current_sound = sound
                self._sound_repeat = repeat
                self._sound_device.write(sound)

    def _cancel_sound(self):
        self._run_sound(0, 1)

    def _encode_motoring_packet(self, address):
        result = "10"
        with self._thread_lock:
            left_wheel = Util.round(self._left_wheel * 1.27)
            if left_wheel > 127: left_wheel = 127
            elif left_wheel < -127: left_wheel = -127
            right_wheel = Util.round(self._right_wheel * 1.27)
            if right_wheel > 127: right_wheel = 127
            elif right_wheel < -127: right_wheel = -127
            if left_wheel == 0 and right_wheel == 0:
                if self._wheel_move:
                    self._wheel_move_count += 1
                    if self._wheel_move_count > 5:
                        self._wheel_move = False
            else:
                self._wheel_move = True
                self._wheel_move_count = 0
            result += self._to_hex(left_wheel)
            result += self._to_hex(right_wheel)
            self._wheel_pulse = self._pulse
            if self._pulse_written:
                if self._wheel_pulse != 0 or self._wheel_pulse_prev != 0:
                    self._wheel_id = (self._wheel_id % 255) + 1
                self._wheel_count = 0
                if self._wheel_pulse > 0:
                    self._wheel_event = 1
                else:
                    self._wheel_event = 0
                self._wheel_pulse_prev = self._wheel_pulse
                self._pulse_written = False
            result += self._to_hex(self._wheel_id)
            result += self._to_hex2(self._wheel_pulse)
            result += self._to_hex(self._mic_led_red)
            result += self._to_hex(self._mic_led_green)
            result += self._to_hex(self._mic_led_blue)
            result += self._to_hex(self._left_eye_red)
            result += self._to_hex(self._left_eye_green)
            result += self._to_hex(self._left_eye_blue)
            result += self._to_hex(self._right_eye_red)
            result += self._to_hex(self._right_eye_green)
            result += self._to_hex(self._right_eye_blue)
            temp = self._get_sound(self._sound)
            if self._sound_written:
                self._sound_count = 0
                if temp > 0:
                    self._sound_flag ^= 0x80
                    self._sound_event = 1
                else:
                    self._sound_flag = 0
                    self._sound_event = 0
                self._sound_written = False
            temp |= self._sound_flag
            result += self._to_hex(temp)
            result += self._to_hex(self._note)
            result += self._to_hex2(Util.round(self._buzzer * 10))
        result += "00-"
        result += address
        result += "\r"
        return result

    def _calc_battery_perc(self, value):
        if value >= 4.11: return 100
        if value >= 3.59: return (100 - 90) * (value - 3.59) / (4.11 - 3.59) + 90
        if value >= 3.55: return (90 - 10) * (value - 3.55) / (3.59 - 3.55) + 10
        if value >= 3.53: return 10 * (value - 3.53) / (3.55 - 3.53)
        return 0

    def _decode_sensory_packet(self, packet):
        packet = str(packet)
        value = int(packet[0:1], 16)
        if value != 1: return False

        value = int(packet[1:2], 16)
        value = (value >> 2) & 0x01
        if value != self._event_tap_id:
            self._tap_device._put_empty(self._event_tap_id != -1 and self._wheel_move == False)
            self._event_tap_id = value
        value = int(packet[36:38], 16)
        value -= 0x100
        self._signal_strength_device._put(value)
        value = int(packet[2:4], 16)
        self._left_proximity_device._put(value)
        value = int(packet[4:6], 16)
        self._right_proximity_device._put(value)
        if self._acceleration_count < 10:
            self._acceleration_count += 1
        else:
            self._acceleration_index %= 10
            self._acceleration_sum_x -= self._acceleration_x[self._acceleration_index]
            self._acceleration_sum_y -= self._acceleration_y[self._acceleration_index]
            self._acceleration_sum_z -= self._acceleration_z[self._acceleration_index]
        value = int(packet[16:18], 16)
        if value > 0x7f: value -= 0x100
        value *= 64
        self._acceleration_sum_x += value
        self._acceleration_x[self._acceleration_index] = value
        value = int(packet[18:20], 16)
        if value > 0x7f: value -= 0x100
        value *= 64
        self._acceleration_sum_y += value
        self._acceleration_y[self._acceleration_index] = value
        value = int(packet[20:22], 16)
        if value > 0x7f: value -= 0x100
        value *= 64
        self._acceleration_sum_z += value
        self._acceleration_z[self._acceleration_index] = value
        self._acceleration_index += 1
        acc_x = int(self._acceleration_sum_x / self._acceleration_count)
        acc_y = int(self._acceleration_sum_y / self._acceleration_count)
        acc_z = int(self._acceleration_sum_z / self._acceleration_count)
        self._acceleration_device._put_at(0, acc_x)
        self._acceleration_device._put_at(1, acc_y)
        self._acceleration_device._put_at(2, acc_z)
        if acc_z < 2048 and acc_x > 2048 and acc_y > -1024 and acc_y < 1024: value = 1
        elif acc_z < 2048 and acc_x < -2048 and acc_y > -1024 and acc_y < 1024: value = -1
        elif acc_z < 2048 and acc_y > 2048 and acc_x > -1024 and acc_x < 1024: value = 2
        elif acc_z < 2048 and acc_y < -2048 and acc_x > -1024 and acc_x < 1024: value = -2
        elif acc_z > 3072 and acc_x > -2048 and acc_x < 2048 and acc_y > -2048 and acc_y < 2048: value = 3
        elif acc_z < -3072 and acc_x > -1024 and acc_x < 1024 and acc_y > -1024 and acc_y < 1024: value = -3
        else: value = 0
        if value != self._event_tilt:
            self._tilt_device._put(value, self._event_tilt != -4)
            self._event_tilt = value
        value = int(packet[6:10], 16)
        self._light_device._put(value)
        value = (int(packet[38:40], 16) + 300) / 100.0
        self._battery_device._put(self._calc_battery_perc(value))
        if value <= 3.55: value = 0
        elif value <= 3.59: value = 1
        else: value = 2
        if value != self._event_battery_state:
            self._battery_state_device._put(value, self._event_battery_state != -1)
            self._event_battery_state = value
        p = int(packet[10:12], 16)
        value = (p >> 7) & 0x01
        if value != self._event_play_touch:
            self._play_touch_device._put(value, self._event_play_touch != -1)
            self._event_play_touch = value
        value = (p >> 6) & 0x01
        if value != self._event_back_touch:
            self._back_touch_device._put(value, self._event_back_touch != -1)
            self._event_back_touch = value
        value = (p >> 2) & 0x01
        if value != self._event_mic_touch:
            self._mic_touch_device._put(value, self._event_mic_touch != -1)
            self._event_mic_touch = value
        value = (p >> 1) & 0x01
        if value != self._event_volume_up_touch:
            self._volume_up_touch_device._put(value, self._event_volume_up_touch != -1)
            self._event_volume_up_touch = value
        value = p & 0x01
        if value != self._event_volume_down_touch:
            self._volume_down_touch_device._put(value, self._event_volume_down_touch != -1)
            self._event_volume_down_touch = value
        self._mic_checker.check(self._event_mic_touch)
        self._volume_up_checker.check(self._event_volume_up_touch)
        self._volume_down_checker.check(self._event_volume_down_touch)
        self._play_checker.check(self._event_play_touch)
        self._back_checker.check(self._event_back_touch)
        value = self._mic_checker.is_clicked()
        if value != self._event_mic_clicked:
            self._mic_clicked_device._put_empty(value)
            self._event_mic_clicked = value
        value = self._volume_up_checker.is_clicked()
        if value != self._event_volume_up_clicked:
            self._volume_up_clicked_device._put_empty(value)
            self._event_volume_up_clicked = value
        value = self._volume_down_checker.is_clicked()
        if value != self._event_volume_down_clicked:
            self._volume_down_clicked_device._put_empty(value)
            self._event_volume_down_clicked = value
        value = self._play_checker.is_clicked()
        if value != self._event_play_clicked:
            self._play_clicked_device._put_empty(value)
            self._event_play_clicked = value
        value = self._back_checker.is_clicked()
        if value != self._event_back_clicked:
            self._back_clicked_device._put_empty(value)
            self._event_back_clicked = value
        value = self._mic_checker.is_long_pressed()
        if value != self._event_mic_long_pressed:
            self._mic_long_pressed_device._put_empty(value)
            self._event_mic_long_pressed = value
        value = self._volume_up_checker.is_long_pressed()
        if value != self._event_volume_up_long_pressed:
            self._volume_up_long_pressed_device._put_empty(value)
            self._event_volume_up_long_pressed = value
        value = self._volume_down_checker.is_long_pressed()
        if value != self._event_volume_down_long_pressed:
            self._volume_down_long_pressed_device._put_empty(value)
            self._event_volume_down_long_pressed = value
        value = self._play_checker.is_long_pressed()
        if value != self._event_play_long_pressed:
            self._play_long_pressed_device._put_empty(value)
            self._event_play_long_pressed = value
        value = self._back_checker.is_long_pressed()
        if value != self._event_back_long_pressed:
            self._back_long_pressed_device._put_empty(value)
            self._event_back_long_pressed = value
        value = self._mic_checker.is_long_long_pressed()
        if value != self._event_mic_long_long_pressed:
            self._mic_long_long_pressed_device._put_empty(value)
            self._event_mic_long_long_pressed = value
        value = self._volume_up_checker.is_long_long_pressed()
        if value != self._event_volume_up_long_long_pressed:
            self._volume_up_long_long_pressed_device._put_empty(value)
            self._event_volume_up_long_long_pressed = value
        value = self._volume_down_checker.is_long_long_pressed()
        if value != self._event_volume_down_long_long_pressed:
            self._volume_down_long_long_pressed_device._put_empty(value)
            self._event_volume_down_long_long_pressed = value
        value = self._play_checker.is_long_long_pressed()
        if value != self._event_play_long_long_pressed:
            self._play_long_long_pressed_device._put_empty(value)
            self._event_play_long_long_pressed = value
        value = self._back_checker.is_long_long_pressed()
        if value != self._event_back_long_long_pressed:
            self._back_long_long_pressed_device._put_empty(value)
            self._event_back_long_long_pressed = value
        value = int(packet[30:32], 16)
        id = (value >> 4) & 0x0f
        oid_mode = value & 0x0f
        if id != self._event_oid_mode_id:
            self._oid_mode_device._put(oid_mode, self._event_oid_mode_id != -1)
            self._event_oid_mode_id = id
        lift = 0
        if oid_mode == 0:
            value = -1
            lift = 1
            self._position_device._put_at(0, -1)
            self._position_device._put_at(1, -1)
            self._orientation_device._put(-200)
        elif oid_mode == 0x0f:
            value = -1
        else:
            if oid_mode == 1:
                value = -1
                self._position_device._put_at(0, int(packet[22:26], 16))
                self._position_device._put_at(1, int(packet[26:30], 16))
            else:
                oid_most = int(packet[22:24], 16)
                oid_high = int(packet[24:26], 16)
                oid_middle = int(packet[26:28], 16)
                oid_low = int(packet[28:30], 16)
                if oid_mode == 2:
                    if oid_most == 0 and (oid_high & 0x40) != 0 and (oid_high & 0x20) == 0:
                        value = ((oid_high & 0x03) << 16) | ((oid_middle & 0xff) << 8) | (oid_low & 0xff)
                        if value < 0x010000:
                            pass
                        elif value < 0x03fff0:
                            value = -1
                        elif value > 0x03fffb and value < 0x040000:
                            value = -1
                    else:
                        value = -2
                elif oid_mode == 3:
                    if (oid_most & 0xf0) == 0:
                        value = ((oid_most & 0x0f) << 24) | ((oid_high & 0xff) << 16) | ((oid_middle & 0xff) << 8) | (oid_low & 0xff)
                    else:
                        value = -2
                else:
                    value = -2
                if value == -2:
                    if self._event_oid == -2:
                        value = -1
                    else:
                        value = self._event_oid
                self._position_device._put_at(0, -1)
                self._position_device._put_at(1, -1)
            orientation = int(packet[32:36], 16)
            if orientation > 180: orientation -= 360
            self._orientation_device._put(orientation)
        if value != self._event_oid:
            self._oid_device._put(value, self._event_oid != -2)
            self._event_oid = value
        if lift != self._event_lift:
            self._lift_device._put(lift, self._event_lift != -1)
            self._event_lift = lift
        value = int(packet[12:16], 16)
        if value != self._event_pulse_count:
            self._pulse_count_device._put(value, self._event_pulse_count != -1)
            self._event_pulse_count = value
        state = (p >> 4) & 0x03
        if self._wheel_event == 1:
            if state == 2:
                self._wheel_count += 1
                if self._wheel_count > 5: self._wheel_event = 2
            elif state == 3:
                self._wheel_event = 2
        if self._wheel_event == 2:
            if state != self._wheel_state or self._wheel_count > 5:
                self._wheel_state = state
                self._wheel_state_device._put(state)
                if state == 2:
                    self._wheel_event = 0
                    self._wheel_count = 0
        if self._wheel_event == -1:
            self._wheel_state = state
            self._wheel_state_device._put(2)
            self._wheel_event = 0
            self._wheel_count = 0
        state = (p >> 3) & 0x01
        if self._sound_event == 1:
            if state == 0:
                self._sound_count += 1
                if self._sound_count > 5: self._sound_event = 2
            else:
                self._soudn_event = 2
        if self._sound_event == 2:
            if state != self._sound_state or self._sound_count > 5:
                self._sound_state = state
                if state == 0:
                    self._sound_event = 0
                    self._sound_count = 0
                    if self._current_sound > 0:
                        if self._sound_repeat < 0:
                            self._run_sound(self._current_sound, -1)
                        elif self._sound_repeat > 1:
                            self._sound_repeat -= 1
                            self._run_sound(self._current_sound, self._sound_repeat)
                        else:
                            self._current_sound = 0
                            self._sound_repeat = 1
                            self._sound_state_device._put(state)
                    else:
                        self._current_sound = 0
                        self._sound_repeat = 1
        return True

    def _receive(self, connector):
        if connector:
            packet = connector.read()
            if packet:
                if self._decode_sensory_packet(packet):
                    if self._ready == False:
                        self._ready = True
                        Runner.register_checked()
                    self._notify_sensory_device_data_changed()
                return True
        return False

    def _send(self, connector):
        if connector:
            packet = self._encode_motoring_packet(connector.get_address())
            connector.write(packet)


# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

from roboid.runner import Runner
from roboid.util import Util
from roboid.model import Robot


class TurtleLoader(object):
    def __init__(self, param):
        self._param = param

    def load(self, serial, address):
        loaded = False
        count = 0
        while True:
            try:
                packet = self._read(serial)
                if packet:
                    if loaded:
                        if packet[0] == "B" and packet[1] == "0":
                            self._parse(packet)
                            self._write(serial, "7000000000000000000000000000000000004400-" + address + "\r")
                        elif packet[0] == "1":
                            return
                    else:
                        if packet[0] == "B" and packet[1] == "0":
                            self._parse(packet)
                            self._write(serial, "7000000000000000000000000000000000004400-" + address + "\r")
                            loaded = True
                        else:
                            if count % 10 == 0:
                                self._write(serial, "B000000000000000000000000000000000004400-" + address + "\r")
                            count += 1
            except:
                pass

    def _read(self, serial):
        try:
            line = bytearray()
            terminator = ord("\r")
            while True:
                c = serial.read()[0]
                line.append(c)
                if c == terminator: break
            return line.decode("utf-8")
        except:
            return ""

    def _write(self, serial, packet):
        try:
            serial.write(packet.encode())
        except:
            pass

    def _parse(self, packet):
        param = self._param
        packet = str(packet)
        value = int(packet[2:4], 16)
        if value > 0x7f: value -= 0x100
        param[0] = value + 1
        value = int(packet[4:6], 16)
        if value > 0x7f: value -= 0x100
        param[1] = value + 1
        value = int(packet[6:8], 16)
        if value > 0x7f: value -= 0x100
        param[2] = value + 1
        value = int(packet[8:10], 16)
        if value > 0x7f: value -= 0x100
        param[3] = value + 1
        value = int(packet[10:14], 16)
        if value > 0x7fff: value -= 0x10000
        distance = 2.3125 + (value + 1) / 10000.0
        if distance <= 0: param[4] = 2.3125
        else: param[4] = distance


class Turtle(Robot):
    ID = "kr.robomation.physical.turtle"

    LEFT_WHEEL = 0x00900000
    RIGHT_WHEEL = 0x00900001
    LED = 0x00900002
    BUZZER = 0x00900003
    PULSE = 0x00900004
    NOTE = 0x00900005
    SOUND = 0x00900006
    LINE_TRACER_MODE = 0x00900007
    LINE_TRACER_GAIN = 0x00900008
    LINE_TRACER_SPEED = 0x00900009
    LAMP = 0x0090000a
    LOCK = 0x0090000b

    SIGNAL_STRENGTH = 0x0090000d
    COLOR = 0x0090000e
    FLOOR = 0x0090000f
    ACCELERATION = 0x00900010
    TEMPERATURE = 0x00900011
    BUTTON = 0x00900012
    CLICKED = 0x00900013
    DOUBLE_CLICKED = 0x00900014
    LONG_PRESSED = 0x00900015
    COLOR_NUMBER = 0x00900016
    COLOR_PATTERN = 0x00900017
    PULSE_COUNT = 0x00900018
    WHEEL_STATE = 0x00900019
    SOUND_STATE = 0x0090001a
    LINE_TRACER_STATE = 0x0090001b
    TILT = 0x0090001c
    BATTERY_STATE = 0x0090001d

    COLOR_NONE = -1
    COLOR_BLACK = 0
    COLOR_RED = 1
    COLOR_ORANGE = 2
    COLOR_YELLOW = 3
    COLOR_GREEN = 4
    COLOR_SKY_BLUE = 5
    COLOR_CYAN = 5
    COLOR_BLUE = 6
    COLOR_PURPLE = 7
    COLOR_MAGENTA = 7
    COLOR_WHITE = 8

    COLOR_NAME_OFF = "off"
    COLOR_NAME_BLACK = "black"
    COLOR_NAME_RED = "red"
    COLOR_NAME_ORANGE = "orange"
    COLOR_NAME_YELLOW = "yellow"
    COLOR_NAME_GREEN = "green"
    COLOR_NAME_SKY_BLUE = "sky blue"
    COLOR_NAME_BLUE = "blue"
    COLOR_NAME_VIOLET = "violet"
    COLOR_NAME_PURPLE = "purple"
    COLOR_NAME_WHITE = "white"

    LINE_TRACER_MODE_OFF = 0
    LINE_TRACER_MODE_BLACK = 10
    LINE_TRACER_MODE_RED = 11
    LINE_TRACER_MODE_GREEN = 13
    LINE_TRACER_MODE_BLUE = 15
    LINE_TRACER_MODE_ANY = 17
    LINE_TRACER_MODE_TURN_LEFT = 20
    LINE_TRACER_MODE_TURN_RIGHT = 30
    LINE_TRACER_MODE_CROSS = 40
    LINE_TRACER_MODE_UTURN = 50
    LINE_TRACER_MODE_BLACK_UNTIL_RED = 61
    LINE_TRACER_MODE_BLACK_UNTIL_YELLOW = 62
    LINE_TRACER_MODE_BLACK_UNTIL_GREEN = 63
    LINE_TRACER_MODE_BLACK_UNTIL_CYAN = 64
    LINE_TRACER_MODE_BLACK_UNTIL_BLUE = 65
    LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA = 66
    LINE_TRACER_MODE_BLACK_UNTIL_ANY = 67
    LINE_TRACER_MODE_RED_UNTIL_BLACK = 71
    LINE_TRACER_MODE_GREEN_UNTIL_BLACK = 73
    LINE_TRACER_MODE_BLUE_UNTIL_BLACK = 75
    LINE_TRACER_MODE_ANY_UNTIL_BLACK = 77

    NOTE_OFF = 0
    NOTE_A_0 = 1
    NOTE_A_SHARP_0 = 2
    NOTE_B_FLAT_0 = 2
    NOTE_B_0 = 3
    NOTE_C_1 = 4
    NOTE_C_SHARP_1 = 5
    NOTE_D_FLAT_1 = 5
    NOTE_D_1 = 6
    NOTE_D_SHARP_1 = 7
    NOTE_E_FLAT_1 = 7
    NOTE_E_1 = 8
    NOTE_F_1 = 9
    NOTE_F_SHARP_1 = 10
    NOTE_G_FLAT_1 = 10
    NOTE_G_1 = 11
    NOTE_G_SHARP_1 = 12
    NOTE_A_FLAT_1 = 12
    NOTE_A_1 = 13
    NOTE_A_SHARP_1 = 14
    NOTE_B_FLAT_1 = 14
    NOTE_B_1 = 15
    NOTE_C_2 = 16
    NOTE_C_SHARP_2 = 17
    NOTE_D_FLAT_2 = 17
    NOTE_D_2 = 18
    NOTE_D_SHARP_2 = 19
    NOTE_E_FLAT_2 = 19
    NOTE_E_2 = 20
    NOTE_F_2 = 21
    NOTE_F_SHARP_2 = 22
    NOTE_G_FLAT_2 = 22
    NOTE_G_2 = 23
    NOTE_G_SHARP_2 = 24
    NOTE_A_FLAT_2 = 24
    NOTE_A_2 = 25
    NOTE_A_SHARP_2 = 26
    NOTE_B_FLAT_2 = 26
    NOTE_B_2 = 27
    NOTE_C_3 = 28
    NOTE_C_SHARP_3 = 29
    NOTE_D_FLAT_3 = 29
    NOTE_D_3 = 30
    NOTE_D_SHARP_3 = 31
    NOTE_E_FLAT_3 = 31
    NOTE_E_3 = 32
    NOTE_F_3 = 33
    NOTE_F_SHARP_3 = 34
    NOTE_G_FLAT_3 = 34
    NOTE_G_3 = 35
    NOTE_G_SHARP_3 = 36
    NOTE_A_FLAT_3 = 36
    NOTE_A_3 = 37
    NOTE_A_SHARP_3 = 38
    NOTE_B_FLAT_3 = 38
    NOTE_B_3 = 39
    NOTE_C_4 = 40
    NOTE_C_SHARP_4 = 41
    NOTE_D_FLAT_4 = 41
    NOTE_D_4 = 42
    NOTE_D_SHARP_4 = 43
    NOTE_E_FLAT_4 = 43
    NOTE_E_4 = 44
    NOTE_F_4 = 45
    NOTE_F_SHARP_4 = 46
    NOTE_G_FLAT_4 = 46
    NOTE_G_4 = 47
    NOTE_G_SHARP_4 = 48
    NOTE_A_FLAT_4 = 48
    NOTE_A_4 = 49
    NOTE_A_SHARP_4 = 50
    NOTE_B_FLAT_4 = 50
    NOTE_B_4 = 51
    NOTE_C_5 = 52
    NOTE_C_SHARP_5 = 53
    NOTE_D_FLAT_5 = 53
    NOTE_D_5 = 54
    NOTE_D_SHARP_5 = 55
    NOTE_E_FLAT_5 = 55
    NOTE_E_5 = 56
    NOTE_F_5 = 57
    NOTE_F_SHARP_5 = 58
    NOTE_G_FLAT_5 = 58
    NOTE_G_5 = 59
    NOTE_G_SHARP_5 = 60
    NOTE_A_FLAT_5 = 60
    NOTE_A_5 = 61
    NOTE_A_SHARP_5 = 62
    NOTE_B_FLAT_5 = 62
    NOTE_B_5 = 63
    NOTE_C_6 = 64
    NOTE_C_SHARP_6 = 65
    NOTE_D_FLAT_6 = 65
    NOTE_D_6 = 66
    NOTE_D_SHARP_6 = 67
    NOTE_E_FLAT_6 = 67
    NOTE_E_6 = 68
    NOTE_F_6 = 69
    NOTE_F_SHARP_6 = 70
    NOTE_G_FLAT_6 = 70
    NOTE_G_6 = 71
    NOTE_G_SHARP_6 = 72
    NOTE_A_FLAT_6 = 72
    NOTE_A_6 = 73
    NOTE_A_SHARP_6 = 74
    NOTE_B_FLAT_6 = 74
    NOTE_B_6 = 75
    NOTE_C_7 = 76
    NOTE_C_SHARP_7 = 77
    NOTE_D_FLAT_7 = 77
    NOTE_D_7 = 78
    NOTE_D_SHARP_7 = 79
    NOTE_E_FLAT_7 = 79
    NOTE_E_7 = 80
    NOTE_F_7 = 81
    NOTE_F_SHARP_7 = 82
    NOTE_G_FLAT_7 = 82
    NOTE_G_7 = 83
    NOTE_G_SHARP_7 = 84
    NOTE_A_FLAT_7 = 84
    NOTE_A_7 = 85
    NOTE_A_SHARP_7 = 86
    NOTE_B_FLAT_7 = 86
    NOTE_B_7 = 87
    NOTE_C_8 = 88

    NOTE_NAME_C = "C"
    NOTE_NAME_C_SHARP = "C#"
    NOTE_NAME_D_FLAT = "Db"
    NOTE_NAME_D = "D"
    NOTE_NAME_D_SHARP = "D#"
    NOTE_NAME_E_FLAT = "Eb"
    NOTE_NAME_E = "E"
    NOTE_NAME_F = "F"
    NOTE_NAME_F_SHARP = "F#"
    NOTE_NAME_G_FLAT = "Gb"
    NOTE_NAME_G = "G"
    NOTE_NAME_G_SHARP = "G#"
    NOTE_NAME_A_FLAT = "Ab"
    NOTE_NAME_A = "A"
    NOTE_NAME_A_SHARP = "A#"
    NOTE_NAME_B_FLAT = "Bb"
    NOTE_NAME_B = "B"

    SOUND_OFF = 0
    SOUND_BEEP = 1
    SOUND_RANDOM = 2
    SOUND_RANDOM_BEEP = 2
    SOUND_SIREN = 3
    SOUND_ENGINE = 4
    SOUND_ROBOT = 5
    SOUND_MARCH = 6
    SOUND_BIRTHDAY = 7
    SOUND_DIBIDIBIDIP = 8
    SOUND_GOOD_JOB = 9

    SOUND_NAME_OFF = "off"
    SOUND_NAME_BEEP = "beep"
    SOUND_NAME_RANDOM_BEEP = "random beep"
    SOUND_NAME_SIREN = "siren"
    SOUND_NAME_ENGINE = "engine"
    SOUND_NAME_ROBOT = "robot"
    SOUND_NAME_MARCH = "march"
    SOUND_NAME_BIRTHDAY = "birthday"
    SOUND_NAME_DIBIDIBIDIP = "dibidibidip"
    SOUND_NAME_GOOD_JOB = "good job"

    TILT_FORWARD = 1
    TILT_BACKWARD = -1
    TILT_LEFT = 2
    TILT_RIGHT = -2
    TILT_FLIP = 3
    TILT_NOT = -3

    BATTERY_NORMAL = 2
    BATTERY_LOW = 1
    BATTERY_EMPTY = 0

    _COLOR2RGB = {
        "off": (0, 0, 0),
        "red": (255, 0, 0),
        "orange": (255, 63, 0),
        "yellow": (255, 255, 0),
        "green": (0, 255, 0),
        "sky_blue": (0, 255, 255),
        "skyblue": (0, 255, 255),
        "sky blue": (0, 255, 255),
        "cyan": (0, 255, 255),
        "blue": (0, 0, 255),
        "violet": (63, 0, 255),
        "purple": (255, 0, 255),
        "magenta": (255, 0, 255),
        "white": (255, 255, 255)
    }
    _COLORS = {
        "black": 0,
        "red": 1,
        "orange": 2,
        "yellow": 3,
        "green": 4,
        "sky_blue": 5,
        "skyblue": 5,
        "sky blue": 5,
        "cyan": 5,
        "blue": 6,
        "purple": 7,
        "magenta": 7,
        "white": 8
    }
    _NOTES = {
        "c": NOTE_C_1,
        "c#": NOTE_C_SHARP_1,
        "db": NOTE_D_FLAT_1,
        "d": NOTE_D_1,
        "d#": NOTE_D_SHARP_1,
        "eb": NOTE_E_FLAT_1,
        "e": NOTE_E_1,
        "f": NOTE_F_1,
        "f#": NOTE_F_SHARP_1,
        "gb": NOTE_G_FLAT_1,
        "g": NOTE_G_1,
        "g#": NOTE_G_SHARP_1,
        "ab": NOTE_A_FLAT_1,
        "a": NOTE_A_1,
        "a#": NOTE_A_SHARP_1,
        "bb": NOTE_B_FLAT_1,
        "b": NOTE_B_1
    }
    _SOUNDS = {
        "off": 0,
        "beep": 1,
        "random beep": 2,
        "random_beep": 2,
        "siren": 3,
        "engine": 4,
        "robot": 5,
        "march": 6,
        "birthday": 7,
        "dibidibidip": 8,
        "good job": 9,
        "good_job": 9
    }
    _robots = {}

    def __init__(self, index=0, port_name=None):
        if isinstance(index, str):
            index = 0
            port_name = index
        if index in Turtle._robots:
            robot = Turtle._robots[index]
            if robot: robot.dispose()
        Turtle._robots[index] = self
        super(Turtle, self).__init__(Turtle.ID, "Turtle", index)
        self._param = [0, 0, 0, 0, 2.3125]
        self._bpm = 60
        self._init(port_name)

    def dispose(self):
        self._roboid._dispose()
        Runner.unregister_robot(self)

    def reset(self):
        self._bpm = 60
        self._roboid._reset()

    def _init(self, port_name):
        from roboid.turtle_roboid import TurtleRoboid
        self._roboid = TurtleRoboid(self.get_index())
        self._add_roboid(self._roboid)
        Runner.register_robot(self)
        Runner.start()
        loader = TurtleLoader(self._param)
        self._roboid._init(loader, port_name)

    def find_device_by_id(self, device_id):
        return self._roboid.find_device_by_id(device_id)

    def _request_motoring_data(self):
        self._roboid._request_motoring_data()

    def _update_sensory_device_state(self):
        self._roboid._update_sensory_device_state()

    def _update_motoring_device_state(self):
        self._roboid._update_motoring_device_state()

    def _notify_sensory_device_data_changed(self):
        self._roboid._notify_sensory_device_data_changed()

    def _notify_motoring_device_data_changed(self):
        self._roboid._notify_motoring_device_data_changed()

    def cm_to_pulse(self, cm):
        if isinstance(cm, (int, float)):
            return Util.round(cm * 106.8)
        else:
            return 0

    def calc_turn_left_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * (1570 + self._param[2]) / 360.0)
        else:
            return 0

    def calc_turn_right_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * (1570 + self._param[0]) / 360.0)
        else:
            return 0

    def calc_pivot_left_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * (3140 + self._param[3]) / 360.0)
        else:
            return 0

    def calc_pivot_right_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * (3140 + self._param[1]) / 360.0)
        else:
            return 0

    def calc_inner_velocity(self, outer_velocity, radius):
        if isinstance(outer_velocity, (int, float)) and isinstance(radius, (int, float)):
            return outer_velocity * (radius - self._param[4]) / (radius + self._param[4])
        else:
            return 0

    def calc_circle_left_pulse(self, degree, radius):
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)):
            return Util.round(degree * (1570 + self._param[2]) * (radius + self._param[4]) / 360.0 / self._param[4])
        else:
            return 0

    def calc_circle_right_pulse(self, degree, radius):
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)):
            return Util.round(degree * (1570 + self._param[0]) * (radius + self._param[4]) / 360.0 / self._param[4])
        else:
            return 0

    def wheels(self, left_velocity, right_velocity=None):
        self.write(Turtle.PULSE, 0)
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        if isinstance(left_velocity, (int, float)):
            self.write(Turtle.LEFT_WHEEL, left_velocity)
        if isinstance(right_velocity, (int, float)):
            self.write(Turtle.RIGHT_WHEEL, right_velocity)
        else:
            if isinstance(left_velocity, (int, float)):
                self.write(Turtle.RIGHT_WHEEL, left_velocity)

    def left_wheel(self, velocity):
        self.write(Turtle.PULSE, 0)
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        if isinstance(velocity, (int, float)):
            self.write(Turtle.LEFT_WHEEL, velocity)

    def right_wheel(self, velocity):
        self.write(Turtle.PULSE, 0)
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        if isinstance(velocity, (int, float)):
            self.write(Turtle.RIGHT_WHEEL, velocity)

    def stop(self):
        self.write(Turtle.PULSE, 0)
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        self.write(Turtle.LEFT_WHEEL, 0)
        self.write(Turtle.RIGHT_WHEEL, 0)

    def _evaluate_wheel_state(self):
        return self.e(Turtle.WHEEL_STATE) and self.read(Turtle.WHEEL_STATE) == 0

    def _motion(self, pulse, left_velocity, right_velocity):
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        if pulse > 0:
            self.write(Turtle.LEFT_WHEEL, left_velocity)
            self.write(Turtle.RIGHT_WHEEL, right_velocity)
            self.write(Turtle.PULSE, pulse)
            Runner.wait_until(self._evaluate_wheel_state)
            self.write(Turtle.LEFT_WHEEL, 0)
            self.write(Turtle.RIGHT_WHEEL, 0)
        else:
            self.write(Turtle.LEFT_WHEEL, 0)
            self.write(Turtle.RIGHT_WHEEL, 0)
            self.write(Turtle.PULSE, 0)

    def move_forward(self, cm=6, velocity=50):
        if isinstance(cm, (int, float)) and isinstance(velocity, (int, float)):
            if cm < 0:
                cm = -cm
                velocity = -velocity
            self._motion(self.cm_to_pulse(cm), velocity, velocity)

    def move_backward(self, cm=6, velocity=50):
        if isinstance(cm, (int, float)) and isinstance(velocity, (int, float)):
            if cm < 0:
                cm = -cm
                velocity = -velocity
            self._motion(self.cm_to_pulse(cm), -velocity, -velocity)

    def turn_left(self, degree=90, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree
                velocity = -velocity
            self._motion(self.calc_turn_left_pulse(degree), -velocity, velocity)

    def turn_right(self, degree=90, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree
                velocity = -velocity
            self._motion(self.calc_turn_right_pulse(degree), velocity, -velocity)

    def pivot_left(self, degree, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree
                velocity = -velocity
            self._motion(self.calc_pivot_left_pulse(degree), 0, velocity)

    def pivot_right(self, degree, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree
                velocity = -velocity
            self._motion(self.calc_pivot_right_pulse(degree), velocity, 0)

    def circle_left(self, degree, radius, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: degree = 0
            if degree < 0:
                degree = -degree
                velocity = -velocity
            left_velocity = self.calc_inner_velocity(velocity, radius)
            pulse = self.calc_circle_left_pulse(degree, radius)
            self._motion(pulse, left_velocity, velocity)

    def swing_left(self, degree, radius, velocity=50):
        self.circle_left(degree, radius, velocity)

    def circle_right(self, degree, radius, velocity=50):
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: degree = 0
            if degree < 0:
                degree = -degree
                velocity = -velocity
            right_velocity = self.calc_inner_velocity(velocity, radius)
            pulse = self.calc_circle_right_pulse(degree, radius)
            self._motion(pulse, velocity, right_velocity)

    def swing_right(self, degree, radius, velocity=50):
        self.circle_right(degree, radius, velocity)

    def _motion_sec(self, sec, left_velocity, right_velocity):
        self.write(Turtle.PULSE, 0)
        self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)
        if sec < 0:
            sec = -sec
            left_velocity = -left_velocity
            right_velocity = -right_velocity
        if sec > 0:
            self.write(Turtle.LEFT_WHEEL, left_velocity)
            self.write(Turtle.RIGHT_WHEEL, right_velocity)
            Runner.wait(sec * 1000)
        self.write(Turtle.LEFT_WHEEL, 0)
        self.write(Turtle.RIGHT_WHEEL, 0)

    def move_forward_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, velocity, velocity)

    def move_backward_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, -velocity, -velocity)

    def turn_left_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, -velocity, velocity)

    def turn_right_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, velocity, -velocity)

    def pivot_left_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, 0, velocity)

    def pivot_right_sec(self, sec, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            self._motion_sec(sec, velocity, 0)

    def circle_left_sec(self, sec, radius, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: sec = 0
            left_velocity = self.calc_inner_velocity(velocity, radius)
            self._motion_sec(sec, left_velocity, velocity)

    def swing_left_sec(self, sec, radius, velocity=50):
        self.circle_left_sec(sec, radius, velocity)

    def circle_right_sec(self, sec, radius, velocity=50):
        if isinstance(sec, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: sec = 0
            right_velocity = self.calc_inner_velocity(velocity, radius)
            self._motion_sec(sec, velocity, right_velocity)

    def swing_right_sec(self, sec, radius, velocity=50):
        self.circle_right_sec(sec, radius, velocity)

    def move_forward_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), velocity, velocity)

    def move_backward_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), -velocity, -velocity)

    def turn_left_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), -velocity, velocity)

    def turn_right_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), velocity, -velocity)

    def pivot_left_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), 0, velocity)

    def pivot_right_pulse(self, pulse, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            self._motion(Util.round(pulse), velocity, 0)

    def circle_left_pulse(self, pulse, radius, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: pulse = 0
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            left_velocity = self.calc_inner_velocity(velocity, radius)
            self._motion(Util.round(pulse), left_velocity, velocity)

    def swing_left_pulse(self, pulse, radius, velocity=50):
        self.circle_left_pulse(pulse, radius, velocity)

    def circle_right_pulse(self, pulse, radius, velocity=50):
        if isinstance(pulse, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0: pulse = 0
            if pulse < 0:
                pulse = -pulse
                velocity = -velocity
            right_velocity = self.calc_inner_velocity(velocity, radius)
            self._motion(Util.round(pulse), velocity, right_velocity)

    def swing_right_pulse(self, pulse, radius, velocity=50):
        self.circle_right_pulse(pulse, radius, velocity)

    def _evaluate_line_tracer(self):
        return self.e(Turtle.LINE_TRACER_STATE) and self.read(Turtle.LINE_TRACER_STATE) == 0x02

    def line_tracer_mode(self, mode):
        self.write(Turtle.LEFT_WHEEL, 0)
        self.write(Turtle.RIGHT_WHEEL, 0)
        self.write(Turtle.PULSE, 0)
        if isinstance(mode, (int, float)):
            mode = int(mode)
            self.write(Turtle.LINE_TRACER_MODE, mode)
            if mode == 20 or mode == 30 or mode == 40 or mode == 50 or (mode >= 61 and mode <= 67) or mode == 71 or mode == 73 or mode == 75 or mode == 77:
                Runner.wait_until(self._evaluate_line_tracer)
                self.write(Turtle.LINE_TRACER_MODE, Turtle.LINE_TRACER_MODE_OFF)

    def black_line(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK)

    def red_line(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_RED)

    def green_line(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_GREEN)

    def blue_line(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLUE)

    def any_line(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_ANY)

    def black_line_until_red(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_RED)

    def black_line_until_yellow(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_YELLOW)

    def black_line_until_green(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_GREEN)

    def black_line_until_sky_blue(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_CYAN)

    def black_line_until_blue(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_BLUE)

    def black_line_until_purple(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA)

    def black_line_until_any(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLACK_UNTIL_ANY)

    def red_line_until_black(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_RED_UNTIL_BLACK)

    def green_line_until_black(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_GREEN_UNTIL_BLACK)

    def blue_line_until_black(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_BLUE_UNTIL_BLACK)

    def any_line_until_black(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_ANY_UNTIL_BLACK)

    def cross_forward(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_CROSS)

    def cross_left(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_TURN_LEFT)

    def cross_right(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_TURN_RIGHT)

    def cross_uturn(self):
        self.line_tracer_mode(Turtle.LINE_TRACER_MODE_UTURN)

    def line_gain(self, gain):
        if isinstance(gain, (int, float)):
            self.write(Turtle.LINE_TRACER_GAIN, Util.round(gain))

    def line_tracer_gain(self, gain):
        self.line_gain(gain)

    def line_speed(self, speed):
        if isinstance(speed, (int, float)):
            self.write(Turtle.LINE_TRACER_SPEED, Util.round(speed))

    def line_tracer_speed(self, speed):
        self.line_speed(speed)

    def rgb(self, red, green=None, blue=None):
        if isinstance(red, (int, float)):
            red = Util.round(red)
            if isinstance(green, (int, float)) and isinstance(blue, (int, float)):
                green = Util.round(green)
                blue = Util.round(blue)
                self.write(Turtle.LED, 0, red)
                self.write(Turtle.LED, 1, green)
                self.write(Turtle.LED, 2, blue)
            else:
                self.write(Turtle.LED, 0, red)
                self.write(Turtle.LED, 1, red)
                self.write(Turtle.LED, 2, red)

    def led(self, color1, color2=None, color3=None):
        if isinstance(color1, (int, float)):
            color1 = Util.round(color1)
            if isinstance(color2, (int, float)) and isinstance(color3, (int, float)):
                color2 = Util.round(color2)
                color3 = Util.round(color3)
                self.write(Turtle.LED, 0, color1)
                self.write(Turtle.LED, 1, color2)
                self.write(Turtle.LED, 2, color3)
            else:
                self.write(Turtle.LED, 0, color1)
                self.write(Turtle.LED, 1, color1)
                self.write(Turtle.LED, 2, color1)
        elif isinstance(color1, str):
            tmp = color1.lower()
            if tmp in Turtle._COLOR2RGB:
                self.write(Turtle.LED, Turtle._COLOR2RGB[tmp])

    def buzzer(self, hz):
        self.write(Turtle.NOTE, Turtle.NOTE_OFF)
        self._roboid._cancel_sound()
        if isinstance(hz, (int, float)):
            self.write(Turtle.BUZZER, hz)

    def tempo(self, bpm):
        if isinstance(bpm, (int, float)):
            if bpm > 0:
                self._bpm = bpm

    def note(self, pitch, beats=None):
        self.write(Turtle.BUZZER, 0)
        self._roboid._cancel_sound()
        if isinstance(pitch, str) and len(pitch) > 0:
            tmp = pitch.lower()
            if tmp == "off":
                pitch = 0
            else:
                octave = 4
                try:
                    octave = int(tmp[-1])
                    tmp = tmp[:-1]
                except ValueError:
                    pass
                if tmp in Turtle._NOTES:
                    pitch = Turtle._NOTES[tmp] + (octave - 1) * 12
        if isinstance(pitch, (int, float)):
            pitch = int(pitch)
            if isinstance(beats, (int, float)):
                bpm = self._bpm
                if beats > 0 and bpm > 0:
                    if pitch == 0:
                        self.write(Turtle.NOTE, Turtle.NOTE_OFF)
                        Runner.wait(beats * 60 * 1000.0 / bpm)
                    elif pitch > 0:
                        timeout = beats * 60 * 1000.0 / bpm
                        tail = 0
                        if timeout > 100:
                            tail = 100
                        self.write(Turtle.NOTE, pitch)
                        Runner.wait(timeout - tail)
                        self.write(Turtle.NOTE, Turtle.NOTE_OFF)
                        if tail > 0:
                            Runner.wait(tail)
                else:
                    self.write(Turtle.NOTE, Turtle.NOTE_OFF)
            elif pitch >= 0:
                self.write(Turtle.NOTE, pitch)

    def _evaluate_sound(self):
        return self.e(Turtle.SOUND_STATE) and self.read(Turtle.SOUND_STATE) == 0

    def sound(self, sound, repeat=1):
        self.write(Turtle.BUZZER, 0)
        self.write(Turtle.NOTE, Turtle.NOTE_OFF)
        if isinstance(sound, str):
            tmp = sound.lower()
            if tmp in Turtle._SOUNDS:
                sound = Turtle._SOUNDS[tmp]
        if isinstance(sound, (int, float)) and isinstance(repeat, (int, float)):
            sound = int(sound)
            repeat = int(repeat)
            if sound > 0 and repeat != 0:
                self._roboid._run_sound(sound, repeat)
            else:
                self._roboid._cancel_sound()

    def sound_until_done(self, sound, repeat=1):
        self.write(Turtle.BUZZER, 0)
        self.write(Turtle.NOTE, Turtle.NOTE_OFF)
        if isinstance(sound, str):
            tmp = sound.lower()
            if tmp in Turtle._SOUNDS:
                sound = Turtle._SOUNDS[tmp]
        if isinstance(sound, (int, float)) and isinstance(repeat, (int, float)):
            sound = int(sound)
            repeat = int(repeat)
            if sound > 0 and repeat != 0:
                self._roboid._run_sound(sound, repeat)
                Runner.wait_until(self._evaluate_sound)
            else:
                self._roboid._cancel_sound()

    def lamp(self, on):
        if on:
            self.write(Turtle.LAMP, 1)
        else:
            self.write(Turtle.LAMP, 0)

    def lock(self, on):
        if on:
            self.write(Turtle.LOCK, 1)
        else:
            self.write(Turtle.LOCK, 0)

    def signal_strength(self):
        return self.read(Turtle.SIGNAL_STRENGTH)

    def floor(self):
        return self.read(Turtle.FLOOR)

    def acceleration_x(self):
        return self.read(Turtle.ACCELERATION, 0)

    def acceleration_y(self):
        return self.read(Turtle.ACCELERATION, 1)

    def acceleration_z(self):
        return self.read(Turtle.ACCELERATION, 2)

    def temperature(self):
        return self.read(Turtle.TEMPERATURE)

    def button(self):
        return self.read(Turtle.BUTTON)

    def clicked(self):
        return self.e(Turtle.CLICKED)

    def double_clicked(self):
        return self.e(Turtle.DOUBLE_CLICKED)

    def long_pressed(self):
        return self.e(Turtle.LONG_PRESSED)

    def color_number(self):
        return self.read(Turtle.COLOR_NUMBER)

    def is_color(self, color):
        if isinstance(color, str):
            tmp = color.lower()
            if tmp in Turtle._COLORS:
                color = Turtle._COLORS[tmp]
        if isinstance(color, (int, float)):
            return int(color) == self.read(Turtle.COLOR_NUMBER)
        return False

    def is_color_number(self, color):
        return self.is_color(color)

    def color_pattern(self):
        if self.e(Turtle.COLOR_PATTERN):
            pattern = self.read(Turtle.COLOR_PATTERN)
            return (pattern // 10, pattern % 10)
        return None

    def is_color_pattern(self, color1, color2):
        if self.e(Turtle.COLOR_PATTERN):
            if isinstance(color1, str):
                tmp = color1.lower()
                if tmp in Turtle._COLORS:
                    color1 = Turtle._COLORS[tmp]
            if isinstance(color2, str):
                tmp = color2.lower()
                if tmp in Turtle._COLORS:
                    color2 = Turtle._COLORS[tmp]
            if isinstance(color1, (int, float)) and isinstance(color2, (int, float)):
                return int(color1) * 10 + int(color2) == self.read(Turtle.COLOR_PATTERN)
        return False

    def pulse_count(self):
        return self.read(Turtle.PULSE_COUNT)

    def tilt(self):
        return self.read(Turtle.TILT)

    def battery_state(self):
        return self.read(Turtle.BATTERY_STATE)